#! /usr/bin/env python3

import pickle
from bisect import bisect_left,bisect_right
import multiprocessing as mp
from copy import deepcopy
import numpy as np
from dbscan1d.core import DBSCAN1D
import argparse
import networkx as nx
import markov_clustering as mc
import scipy.sparse as sp
import os
from os.path import isfile

def which_chromo(seqlen,seqids,i):
    seqlen = sorted(seqlen)
    pos = bisect_left(seqlen,i)
    if i == seqlen[pos]:
        pos += 1
    return(seqids[pos])

def get_regions(dbs,org,iss,j):
    iss.sort()
    labels = dbs.fit_predict(iss)
    regions = {}
    for enu,ident in enumerate(labels):
        if ident == -1:
            continue
        else:
            if ident not in regions:
                regions[ident] = []
            regions[ident].append(iss[enu])
    res = {}
    for ident,region in regions.items():
        res[ident] = [min(region),max(region)]
    return org,res

def get_elements_and_syntenies_of_regions(global_regions):
    anchors_per_region = {}
    cont_eles = {}
    for org,bib in global_regions.items():
        anchors_per_region[org] = {}
        cont_eles[org] = {}
        for ident,t in sorted(bib.items(), key=lambda x: x[1]):
            anchors_per_region[org][ident] = {}
            start,end = t
            abs_s = start
            abs_e = end
            seqids,seqlen = small_meta[org]
            chromo = which_chromo(seqlen,seqids,start)
            chromo_end = which_chromo(seqlen,seqids,end)
            idx_l = seqids.index(chromo)
            if idx_l > 0:
                le = seqlen[idx_l-1]
            else:
                le = 0
            len_chr = seqlen[idx_l]
            start -= le
            start = max(0,start)
            end -= le
            end = min(len_chr,end)
            
            if org not in coords or chromo not in coords[org]:
                elements = ''
            else:
                elements = sorted(coords[org][chromo])
                starts,ends,names = zip(*elements)
                idx_first_ele = bisect_right(starts,start)
                idx_last_ele = bisect_left(ends,end)
                if idx_last_ele - idx_first_ele > 0:
                    elements = ';'.join(names[idx_first_ele:idx_last_ele])
                else:
                    elements = ''

            anchors = sorted(list(collect[org].keys()))
            idx_first_ele = bisect_right(anchors,abs_s)
            idx_last_ele = bisect_left(anchors,abs_e)
            if idx_last_ele - idx_first_ele > 0:
                for a in anchors[idx_first_ele:idx_last_ele]:
                    anchors_per_region[org][ident][a] = deepcopy(collect[org][a])
            cont_eles[org][ident] = elements

    pw_syn = {}
    G=nx.Graph()
    
    for org,bib in anchors_per_region.items():
        pw_syn[org] = {}
        for ident1,bib2 in bib.items():
            pw_syn[org][ident1] = {}
            for a,bib3 in bib2.items():
                for org2,lj in bib3.items():
                    if org2 in anchors_per_region:
                        for ident2,bib22 in anchors_per_region[org2].items():
                            for j in lj:
                                if j in bib22:
                                    if org2 not in pw_syn[org][ident1]:
                                        pw_syn[org][ident1][org2] = {}
                                    if ident2 not in pw_syn[org][ident1][org2]:
                                        pw_syn[org][ident1][org2][ident2] = 0
                                    pw_syn[org][ident1][org2][ident2] += 1

    G.add_nodes_from([f'{o}$$${x}' for o,x in pw_syn.items()])
    for org,bib in pw_syn.items():
        for id1,bib2 in bib.items():
            for org2,bib3 in bib2.items():
                for id2,no in bib3.items():
                    G.add_edge(f'{org}$$${id1}',f'{org2}$$${id2}',weight=no)

    G_trans = nx.transitive_closure(G,reflexive=None)
    trans_syn = {}

    for c in sorted(nx.connected_components(G_trans), key=len, reverse=True):
        c = list(c)
        if len(c) > 1:
            eles = []
            for i in c:
                org,ident = i.split('$$$')
                ident = int(ident)
                eles.append((org,ident))
            for ele in eles:
                for ele2 in eles:
                    if ele != ele2 and ele[0] != ele2[0]:
                        org1,ident1 = ele
                        org2,ident2 = ele2
                        if org1 in pw_syn and ident1 in pw_syn[org1] and org2 in pw_syn[org1][ident1] and ident2 in pw_syn[org1][ident1][org2]:
                            continue
                        if org1 not in trans_syn:
                            trans_syn[org1] = {}
                        if ident1 not in trans_syn[org1]:
                            trans_syn[org1][ident1] = {}
                        if org2 not in trans_syn[org1][ident1]:
                            trans_syn[org1][ident1][org2] = {}
                        trans_syn[org1][ident1][org2][ident2] = -42

    nodelist = []
    mapping = {}
    c = 0
    for node in G.nodes():
        nodelist.append(node)
        mapping[c] = node
        c += 1
    matrix = nx.to_scipy_sparse_array(G, nodelist=nodelist)
    matrix = sp.csr_matrix(matrix)  # Convert sparse array to sparse matrix for markov_clustering (there was some compatibility issue with a newer version of network and the expected input for mcl clustering)
    result = mc.run_mcl(matrix)
    clusters_mcl = mc.get_clusters(result)

    pw_syn_mcl = {}

    for c in clusters_mcl:
        c = list(c)
        if len(c) > 1:
            eles = []
            for i in c:
                i = mapping[i]
                org,ident = i.split('$$$')
                ident = int(ident)
                eles.append((org,ident))
            for ele in eles:
                for ele2 in eles:
                    if ele != ele2 and ele[0] != ele2[0]:
                        org1,ident1 = ele
                        org2,ident2 = ele2
                        if org1 not in pw_syn_mcl:
                            pw_syn_mcl[org1] = {}
                        if ident1 not in pw_syn_mcl[org1]:
                            pw_syn_mcl[org1][ident1] = {}
                        if org2 not in pw_syn_mcl[org1][ident1]:
                            pw_syn_mcl[org1][ident1][org2] = {}
                        pw_syn_mcl[org1][ident1][org2][ident2] = -21

    return cont_eles,pw_syn,trans_syn,pw_syn_mcl

def get_new_regions_to_eval(margin_to_be_new,org,global_reg,old_reg):
    if len(old_reg) == 0:
        return org,[(s,e) for s,e in global_reg.values()]
    new_regions_to_eval = []
    starts = sorted([x[0] for x in old_reg])
    ends = sorted([x[1] for x in old_reg])
    for ident,region in global_reg.items():
        s,e = region
        idx_s = bisect_left(starts,s)
        idx_e = bisect_right(ends,e)
        if idx_s != idx_e:
            continue
        else:
            if (idx_s == 0 and e < starts[0]) or idx_s == len(old_reg):
                new_regions_to_eval.append((s,e))
            else:
                if e < starts[idx_s] and starts[idx_s] - e > margin_to_be_new and s > ends[idx_s - 1] and s - ends[idx_s - 1] > margin_to_be_new:
                    new_regions_to_eval.append((s-1000,e+1000))

    return org,new_regions_to_eval

def get_coords(coords_to_exclude,aligned_path,query_org,target_orgs,regions,first=0):
    if first == -1:
        regions = list(regions.values())
    if isfile(f'{aligned_path}/{query_org}_with_syn_eval'):
        ad = f'{aligned_path}/{query_org}_with_syn_eval'
    else:
        ad = f'{aligned_path}/{query_org}'
    with open(f'{aligned_path}/{query_org}','rb') as f:
        aligned = pickle.load(f)
    iss = sorted(aligned.keys())
    matching_anchors = {}
    for s,e in regions:
        local_iss = iss[bisect_left(iss,s):min(len(iss), bisect_left(iss,e))]
        for i in local_iss:
            if first == 0 and query_org in coords_to_exclude:
                valid = 1
                for sx,ex in coords_to_exclude[query_org]:
                    if i >= sx and i <= ex:
                        valid = 0
                        break
                    elif i < sx and sx - i > 100:
                        valid = 0
                        break
                    elif i > ex and i - ex > 100:
                        valid = 0
                        break
                if valid == 0:
                    continue
            bib = aligned[i]
            for org in target_orgs:
                if org in bib['matches'] and bib['matches'][org]['meta']['multiple matches out of tolerance range'] != 1:
                    if 'matches' in bib['matches'][org] and len(bib['matches'][org]['matches']) > 0:
                        matches = list(bib['matches'][org]['matches'].keys())
                        if i not in matching_anchors:
                            matching_anchors[i] = {}
                        if org not in matching_anchors[i]:
                            matching_anchors[i][org] = set()
                        matching_anchors[i][org].update(matches)
                    if 'dups_matches' in bib['matches'][org] and len(bib['matches'][org]['dups_matches']) > 1 and 'syntenic' in bib['matches'][org]['dups_matches'] and len(bib['matches'][org]['dups_matches']['syntenic']) > 0:
                        matches = bib['matches'][org]['dups_matches']['syntenic']
                        if i not in matching_anchors:
                            matching_anchors[i] = {}
                        if org not in matching_anchors[i]:
                            matching_anchors[i][org] = set()
                        matching_anchors[i][org].update(matches)
    return query_org,matching_anchors

def reconcile_collect(res,collect):
    for qo,ma in res:
        if qo not in collect:
            collect[qo] = {}
        for i,bib in ma.items():
            if i not in collect[qo]:
                collect[qo][i] = {}
            for to,sj in bib.items():
                if to not in collect[qo][i]:
                    collect[qo][i][to] = set()
                if to not in collect:
                    collect[to] = {}
                for j in sj:
                    if j not in collect[to]:
                        collect[to][j] = {}
                    if qo not in collect[to][j]:
                        collect[to][j][qo] = set()
                    collect[qo][i][to].add(j)
                    collect[to][j][qo].add(i)
    return collect 


def parse_gff_dir(gff_dir):
    """
    Parse all GFF files in directory
    Returns list of tuples: (org, chromo, start, end, strand, name)
    """
    elements = []

    if not os.path.isdir(gff_dir):
        print(f"Error: GFF directory not found: {gff_dir}")
        return elements

    for filename in os.listdir(gff_dir):
        if not filename.endswith('.gff'):
            continue

        org = filename.replace('.gff','')
        gff_file = os.path.join(gff_dir, filename)

        with open(gff_file,'r') as f:
            for line in f:
                if line.startswith('#') or not line.strip():
                    continue

                cols = line.strip().split('\t')
                if len(cols) < 9:
                    continue

                seqid = cols[0]
                feature_type = cols[2]
                start = cols[3]
                end = cols[4]
                strand = cols[6]
                attributes = cols[8]

                # parse name from attributes
                name = None
                for attr in attributes.split(';'):
                    attr = attr.strip()
                    if '=' in attr:
                        key,val = attr.split('=',1)
                        if key in ['protein_id','gene_id','ID','Name']:
                            name = val
                            break

                if name is None:
                    name = f'{org}_{seqid}_{start}_{end}'

                elements.append((org, seqid, start, end, strand, name))

    return elements

if __name__ == "__main__":
    parser= argparse.ArgumentParser(description='Find Syntenic Regions From List of Genetic Elements of Interest',formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--cores', type=int,nargs='?', default=1,help="number of cores to run with")
    parser.add_argument('--iter', type=int,nargs='?', default=3,help="number of iterations to find anchors and thus syntenic regions")
    parser.add_argument('--threshold_new_region', type=int,nargs='?', default=1e6,help="margin of syntenic regions, i.e.: number of nulceotides within newly added anchors are considered to be within a neighbouring syntenic region")
    parser.add_argument('--anchor_scope', type=int,nargs='?', default=5e5,help="number of nucleotides up-and downstream of input genetic elements to be considered for initial anchor search")
    parser.add_argument('--no_anchors_form_region', type=int,nargs='?', default=4,help="minimum number of anchors to form a'syntenic region'")
    parser.add_argument('--anchor_path', type=str,nargs='?', default='../../utils/anchors/')
    parser.add_argument('--small_meta_path', type=str,nargs='?', default='../../utils/small_meta/')
    parser.add_argument('--coords', type=str,nargs='?',help="file with coordinates of elements of interest in format:species*tab*chr*tab*start*tab*end*tab*strand (+/-)*tab*[optional:name]")
    parser.add_argument('--gff', type=str,nargs='?',help="directory with GFF files ({org}.gff). Will extract all features. Can be used with --coords")
    parser.add_argument('--coords_to_exclude', type=str,nargs='?',help="file with coordinates of additional elements to exclude from being considered as anchors (in same format as elements of interest: species*tab*chr*tab*start*tab*end*tab*strand(+/-)*tab*[optional:name]. if you happen to know your genomic environment it may be helpful to exclude known elements which should not be considered as synteny anchors since AncST produces <= 5 percent false positives. so if you e.g. want to be rather conservative, are in an environment with complex duplication histories, are working with rather badly assembled genomes, etc. you might want to exclude those elements. on the other hand, this will generally reduce the resolution of synteny anchors and in almost all cases it is fine to leave this option empty and just consider all anchors around since there are additional 'safety' checks and 'fishy' results will show as inconsistent and/or 'too broad' clustering of the regions considered")
    parser.add_argument('--region_buffer', type=int,nargs='?', default=10000,help="number of nucleotides to extend syntenic region boundaries (buffer/padding to avoid missing genes at edges)")
    args = parser.parse_args()

    if not args.coords and not args.gff:
        print("Error: Must provide either --coords or --gff (or both)")
        exit(1)
    if args.cores:
        no_cores = int(args.cores)
    if args.iter:
        no_iter = int(args.iter)
    if args.threshold_new_region:
        margin_to_be_new = int(args.threshold_new_region)
    if args.anchor_scope:
        margin_anchors = int(args.anchor_scope)
    if args.no_anchors_form_region:
        no_anchors_form_region = int(args.no_anchors_form_region)
    if args.anchor_path:
        anchor_path = args.anchor_path
    if args.small_meta_path:
        small_meta_path = args.small_meta_path
    if args.region_buffer:
        region_buffer = int(args.region_buffer)

    aligned_path = f'{anchor_path}/aligned/'
    candidates_path = f'{anchor_path}/candidates/'

    orgs = []
    if os.path.isfile('orgs'):
        with open('orgs','r') as f:
            for line in f:
                orgs.append(line.strip())
    else:
        orgs = [x for x in os.listdir(f'{anchor_path}/aligned/') if 'syn_eval' not in x]

    target_orgs = orgs

    small_meta = {}
    for org in orgs:
        with open(f'{small_meta_path}/{org}','rb') as f:
            small_meta[org] = pickle.load(f)

    dbs = DBSCAN1D(eps=margin_anchors/3, min_samples=no_anchors_form_region)

    collect = {}
    coords = {}
    coords_to_exclude = {}

    # collect elements from both sources
    all_elements = []

    # parse GFF if provided
    if args.gff:
        gff_elements = parse_gff_dir(args.gff)
        all_elements.extend(gff_elements)
        print(f"Loaded {len(gff_elements)} elements from GFF files")

    # parse coords if provided
    if args.coords:
        name_count = 1
        with open(args.coords) as f:
            for line in f:
                line = line.strip().split()
                if len(line) == 6:
                    org,chromo,start,end,ori,name = line
                elif len(line) == 5:
                    org,chromo,start,end,ori = line
                    name = f'element_on_line_{name_count}'
                    name_count += 1
                else:
                    print(line)
                    print('please provide a coodinates file of form species(name of genome/anchors)\tchr\tstart\tend\tstrand (+/-)\t[optional:name]')
                    exit(0)
                all_elements.append((org,chromo,start,end,ori,name))
        print(f"Loaded {len(all_elements) - (len(gff_elements) if args.gff else 0)} elements from coords file")

    print(f"Total elements to process: {len(all_elements)}")

    # process all elements
    for org,chromo,start,end,ori,name in all_elements:
        start = int(start)
        end = int(end)
        mid = int(start+(end-start)/2)
        seqids,seqlen = small_meta[org]
        if chromo not in seqids:
            print(f'{chromo} not in genome {org}')
            continue
        else:
            print(f'{chromo} in genome {org}')
        idx_l = seqids.index(chromo)
        if idx_l > 0:
            l = seqlen[idx_l-1]
        else:
            l = 0
        len_chr = seqlen[idx_l]
        mid += l
        if org not in collect:
            collect[org] = {}
            coords[org] = {}
        collect[org][mid-10000] = {}
        collect[org][mid] = {}
        collect[org][mid+10000] = {}
        if chromo not in coords[org]:
            coords[org][chromo] = []
        coords[org][chromo].append((start,end,name))
        if org not in coords_to_exclude:
            coords_to_exclude[org] = []
        coords_to_exclude[org].append((start,end))

    if args.coords_to_exclude:
        coords_to_exclude_file = args.coords_to_exclude
        with open(coords_to_exclude_file) as f:
            for line in f:
                line = line.strip().split()
                org,chromo,start,end,ori_gene,name = line
                start = int(start)
                end = int(end)
                seqids,seqlen = small_meta[org]
                idx_l = seqids.index(chromo)
                if idx_l > 0:
                    l = seqlen[idx_l-1]
                else:
                    l = 0
                len_chr = seqlen[idx_l]
                start += l
                end += l
                if org not in coords_to_exclude:
                    coords_to_exclude[org] = []
                coords_to_exclude[org].append((start,end))

    initial_regions = {}
    for org,bib in collect.items():
        initial_regions[org] = []
        for mid in bib:
            initial_regions[org].append((max(0,mid-margin_anchors),mid+margin_anchors))

    print(f'iteration no 1')

    with mp.Pool(processes=no_cores) as p:
        res = p.starmap(get_coords,[(coords_to_exclude,aligned_path,query_org,target_orgs,regions,1) for query_org,regions in initial_regions.items()])

    collect = reconcile_collect(res,collect)

    j = 1
    with mp.Pool(processes=no_cores) as p:
        res = p.starmap(get_regions,[(dbs,org,np.array(list(bib.keys())),j) for org,bib in collect.items()])

    global_regions = {}
    for org,regions in res:
        if len(regions) > 0:
            global_regions[org] = regions

    old_regions = deepcopy(initial_regions)
    for org in list(global_regions.keys()):
        if org not in old_regions:
            old_regions[org] = []

    with mp.Pool(processes=no_cores) as p:
        res = p.starmap(get_new_regions_to_eval,[(margin_to_be_new,org,global_regions[org],old_regions[org]) for org in list(global_regions.keys())])
   
    new_regions = {}
    for org,regions in res:
        new_regions[org] = regions

    total_new = sum([len(v) for v in new_regions.values()])
    print(f'found {total_new} new regions after iteration 1')

    for j in range(2,no_iter+1):

        if len(new_regions) == 0 or sum([len(v) for v in new_regions.values()]) == 0:
            print(f'exiting before iteration no {j} since no new regions found')
            break

        print(f'iteration no {j}')

        with mp.Pool(processes=no_cores) as p:
            res = p.starmap(get_coords,[(coords_to_exclude,aligned_path,query_org,target_orgs,regions) for query_org,regions in new_regions.items()])
        
        collect = reconcile_collect(res,collect)

        old_regions = {}
        for org,bib in global_regions.items():
            old_regions[org] = list(bib.values())

        with mp.Pool(processes=no_cores) as p:
            res = p.starmap(get_regions,[(dbs,org,np.array(list(bib.keys())),j) for org,bib in collect.items()])

        global_regions = {}
        for org,regions in res:
            if len(regions) > 0:
                global_regions[org] = regions

        for org in list(global_regions.keys()):
            if org not in old_regions:
                old_regions[org] = []

        with mp.Pool(processes=no_cores) as p:
            res = p.starmap(get_new_regions_to_eval,[(margin_to_be_new,org,global_regions[org],old_regions[org]) for org in list(global_regions.keys())])

        new_regions = {}
        for org,regions in res:
            new_regions[org] = regions

        total_new = sum([len(v) for v in new_regions.values()])
        print(f'found {total_new} new regions after iteration {j}')
    
    with mp.Pool(processes=no_cores) as p:
        res = p.starmap(get_coords,[(coords_to_exclude,aligned_path,query_org,target_orgs,regions,-1) for query_org,regions in global_regions.items()])
    collect = reconcile_collect(res,collect)

    if len(global_regions) == 0 or sum([len(v) for v in list(global_regions.values())]) == 0:
        print(f'no regions found')
        exit()
    cont_eles,pw_syn,trans_syn,pw_syn_mcl = get_elements_and_syntenies_of_regions(global_regions)

    bp_no = 1

    print('writing final output file')
    with open(f'syntenic_regions_verbose','w') as f:
        for org,bib in global_regions.items():
            f.write(f'==============================================================================================================\n')
            f.write(f'{org}\n')
            for ident,t in sorted(bib.items(), key=lambda x: x[1]):
                start,end = t
                elements = cont_eles[org][ident]
                seqids,seqlen = small_meta[org]
                chromo = which_chromo(seqlen,seqids,start)
                chromo_end = which_chromo(seqlen,seqids,end)
                if chromo_end != chromo:
                    seqlen_chr_start = seqlen[seqids.index(chromo)]
                    save_end = end
                    end = seqlen_chr_start - 1
                idx_l = seqids.index(chromo)
                if idx_l > 0:
                    le = seqlen[idx_l-1]
                else:
                    le = 0
                len_chr = seqlen[idx_l]
                start -= le
                start = max(0,start)
                end -= le
                end = min(len_chr,end)
                if chromo_end != chromo:
                    elements1 = []
                    if org in coords and chromo in coords[org]:
                        for x in coords[org][chromo]:
                            if x[0] >= start and x[1] <= end:
                                elements1.append(x[-1])
                    f.write('-----------------------------------------------------------------------------------------------------------------\n')
                    f.write(f'<<< {org}\tregion {ident}\t{chromo}\t{start}\t{end} (== end of chr) \tPOTENTIAL CHROMOSOMAL BREAKPOINT {bp_no} REGION 1 >>>\n')
                    f.write(f'\telements: {elements1}\n')
                    f.write(f'\tdirectly pairwise syntenic to [1 = after markov chain clustering (strength connections between regions == number of anchor matches); 0 = could not hold up MCL]:\n')
                    if org in pw_syn and ident in pw_syn[org]:
                        for org2,l in pw_syn[org][ident].items():
                            if len(l) == 0:
                                continue
                            l_mcl = []
                            if org in pw_syn_mcl and ident in pw_syn_mcl[org] and org2 in pw_syn_mcl[org][ident]:
                                l_mcl = pw_syn_mcl[org][ident][org2]
                            if len(l) > 1:
                                f.write(f'\t\t{org2} (potential breakpoints as multiple regions match - likely just artificial partition into multiple regions due to actual sequence insertions/deletions or program parameter/anchor availability/assembly artefacts):\n')
                            else:
                                f.write(f'\t\t{org2}:\n')
                            for reg2 in l:
                                if reg2 in l_mcl:
                                    f.write(f'\t\t\tregion {reg2} [1] \n')
                                else:
                                    f.write(f'\t\t\tregion {reg2} [0] \n')
                    f.write(f'\tadditionally transitively pairwise syntenic to:\n')
                    if org in trans_syn and ident in trans_syn[org]:
                        for org2,l in trans_syn[org][ident].items():
                            if len(l) == 0:
                                continue
                            if len(l) > 1:
                                f.write(f'\t\t{org2} (potential breakpoints as multiple regions match - likely just artificial partition into multiple regions due to actual sequence insertions/deletions or program parameter/anchor availability/assembly artefacts):\n')
                            else:
                                f.write(f'\t\t{org2}:\n')
                            for reg2 in l:
                                f.write(f'\t\t\tregion {reg2}\n')
                    f.write('-----------------------------------------------------------------------------------------------------------------\n')
                    start = 1
                    end = save_end
                    idx_l = seqids.index(chromo_end)
                    if idx_l > 0:
                        le = seqlen[idx_l-1]
                    else:
                        le = 0
                    len_chr = seqlen[idx_l]
                    end -= le
                    end = min(len_chr,end)
                    elements2 = []
                    if org in coords and chromo_end in coords[org]:
                        for x in coords[org][chromo_end]:
                            if x[0] >= start and x[1] <= end:
                                elements2.append(x[-1])
                    f.write('-----------------------------------------------------------------------------------------------------------------\n')
                    f.write(f'<<< {org}\tregion {ident}\t{chromo_end}\t{start} (== start of chr) \t{end}\tPOTENTIAL CHROMOSOMAL BREAKPOINT {bp_no} REGION 2 >>>\n')
                    f.write(f'\telements: {elements2}\n')
                    f.write(f'\tdirectly pairwise syntenic to:\n')
                    f.write(f'\tdirectly pairwise syntenic to [1 = after markov chain clustering (strength connections between regions == number of anchor matches); 0 = could not hold up MCL]:\n')
                    if org in pw_syn and ident in pw_syn[org]:
                        for org2,l in pw_syn[org][ident].items():
                            if len(l) == 0:
                                continue
                            l_mcl = []
                            if org in pw_syn_mcl and ident in pw_syn_mcl[org] and org2 in pw_syn_mcl[org][ident]:
                                l_mcl = pw_syn_mcl[org][ident][org2]
                            if len(l) > 1:
                                f.write(f'\t\t{org2} (potential breakpoints as multiple regions match - likely just artificial partition into multiple regions due to actual sequence insertions/deletions or program parameter/anchor availability/assembly artefacts):\n')
                            else:
                                f.write(f'\t\t{org2}:\n')
                            for reg2 in l:
                                if reg2 in l_mcl:
                                    f.write(f'\t\t\tregion {reg2} [1] \n')
                                else:
                                    f.write(f'\t\t\tregion {reg2} [0] \n')
                    f.write(f'\tadditionally transitively pairwise syntenic to:\n')
                    if org in trans_syn and ident in trans_syn[org]:
                        for org2,l in trans_syn[org][ident].items():
                            if len(l) == 0:
                                continue
                            if len(l) > 1:
                                f.write(f'\t\t{org2} (potential breakpoints as multiple regions match - likely just artificial partition into multiple regions due to actual sequence insertions/deletions or program parameter/anchor availability/assembly artefacts):\n')
                            else:
                                f.write(f'\t\t{org2}:\n')
                            for reg2 in l:
                                f.write(f'\t\t\tregion {reg2}\n')
                    f.write('-----------------------------------------------------------------------------------------------------------------\n')
                    bp_no += 1
                else:
                    f.write('-----------------------------------------------------------------------------------------------------------------\n')
                    f.write(f'<<< {org}\tregion {ident}\t{chromo}\t{start}\t{end} >>>\n')
                    f.write(f'\telements: {elements}\n')
                    f.write(f'\tdirectly pairwise syntenic to [1 = after markov chain clustering (strength connections between regions == number of anchor matches); 0 = could not hold up MCL]:\n')
                    if org in pw_syn and ident in pw_syn[org]:
                        for org2,l in pw_syn[org][ident].items():
                            if len(l) == 0:
                                continue
                            l_mcl = []
                            if org in pw_syn_mcl and ident in pw_syn_mcl[org] and org2 in pw_syn_mcl[org][ident]:
                                l_mcl = pw_syn_mcl[org][ident][org2]
                            if len(l) > 1:
                                f.write(f'\t\t{org2} (potential breakpoints as multiple regions match - likely just artificial partition into multiple regions due to actual sequence insertions/deletions or program parameter/anchor availability/assembly artefacts):\n')
                            else:
                                f.write(f'\t\t{org2}:\n')
                            for reg2 in l:
                                if reg2 in l_mcl:
                                    f.write(f'\t\t\tregion {reg2} [1] \n')
                                else:
                                    f.write(f'\t\t\tregion {reg2} [0] \n')
                    f.write(f'\tadditionally transitively pairwise syntenic to:\n')
                    if org in trans_syn and ident in trans_syn[org]:
                        for org2,l in trans_syn[org][ident].items():
                            if len(l) == 0:
                                continue
                            if len(l) > 1:
                                f.write(f'\t\t{org2} (potential breakpoints as multiple regions match - likely just artificial partition into multiple regions due to actual sequence insertions/deletions or program parameter/anchor availability/assembly artefacts):\n')
                            else:
                                f.write(f'\t\t{org2}:\n')
                            for reg2 in l:
                                f.write(f'\t\t\tregion {reg2}\n')
                    f.write('-----------------------------------------------------------------------------------------------------------------\n')
            f.write(f'==============================================================================================================\n')
    with open(f'syntenic_regions_succinct','w') as f:
        for org,bib in global_regions.items():
            for ident,t in sorted(bib.items(), key=lambda x: x[1]):
                start,end = t
                seqids,seqlen = small_meta[org]
                chromo = which_chromo(seqlen,seqids,start)
                chromo_end = which_chromo(seqlen,seqids,end)
                if chromo_end != chromo:
                    seqlen_chr_start = seqlen[seqids.index(chromo)]
                    save_end = end
                    end = seqlen_chr_start - 1
                idx_l = seqids.index(chromo)
                if idx_l > 0:
                    le = seqlen[idx_l-1]
                else:
                    le = 0
                len_chr = seqlen[idx_l] - le
                start -= le
                start -= region_buffer
                start = max(0,start)
                end -= le
                end += region_buffer
                end = min(len_chr,end)
                if chromo_end != chromo:
                    start1 = start
                    end1 = len_chr - 1
                    f.write(f'{org}\t{chromo}\t{start1}\t{end1}\tPOT BP1\n')
                    start2 = 0
                    idx_l = seqids.index(chromo_end)
                    if idx_l > 0:
                        le = seqlen[idx_l-1]
                    else:
                        le = 0
                    len_chr = seqlen[idx_l] - le
                    end2 = save_end - le
                    end2 += region_buffer
                    end2 = min(len_chr,end2)
                    f.write(f'{org}\t{chromo_end}\t{start2}\t{end2}\tPOT BP2\n')
                else:
                    f.write(f'{org}\t{chromo}\t{start}\t{end}\tNO BP\n')



#! /usr/bin/env python3

from bisect import bisect_left
import sys
sys.path.append('../../utils/')
from get_mapping import get_mapping

def overlap(start,end,coords_to_check):
    idx = bisect_left([x[0] for x in coords_to_check],start)
    if idx == len(coords_to_check):
        if coords_to_check[idx-1][1] > start:
            return True
        else:
            return False
    if end > coords_to_check[idx][0]:
        return True
    return False

def write_gff():

    elements = []
    take = {}
    out = open(f'MCScanX_plus_coords_only_relevant_{margin}.gff','w')
    with open('MCScanX.gff') as f:
        for line in f:
            chromo,name,start,end = line.strip().split()
            start = int(start)
            end = int(end)
            if chromo not in filtered_coords_per_chr or end < min_max_per_chr[chromo][0] or start > min_max_per_chr[chromo][1]:
                continue
            if chromo in filtered_coords_per_chr and overlap(start,end,filtered_coords_per_chr[chromo]):
                continue
            else:
                out.write(f'{chromo}\t{name}\t{start}\t{end}\n')
                if chromo not in take:
                    take[chromo] = []
                take[chromo].append(name)
    for ele in coords:
        chromo,id_seq,start,end = ele
        out.write(f'{chromo}\t{id_seq}\t{start}\t{end}\n')
        elements.append((id_seq,chromo))
    out.close()

    return elements,take

def write_homology(elements,take):

    homology = {}
    out = open(f'MCScanX_plus_coords_only_relevant_{margin}.homology','w')
    with open('MCScanX.homology') as f:
        for line in f:
            id1,id2,score = line.strip().split()
            chromo1 = id1.split('ele')[0]
            chromo2 = id2.split('ele')[0]
            if chromo1 not in take or id1 not in take[chromo1] or chromo2 not in take or id2 not in take[chromo2]:
                continue
            out.write(f'{id1}\t{id2}\t{score}\n')
    for enu,t in enumerate(elements):
        id1,org1 = t
        for t2 in elements[enu+1:]:
            id2,org2 = t2
            if org1 == org2:
                continue
            out.write(f'{id1}\t{id2}\t9999999999\n')
    out.close()

    return 0

if __name__ == "__main__":
    
    margin = int(sys.argv[1])
    org_mapping,chr_mapping = get_mapping()
   
    coords = []
    coords_per_chr = {}
    with open('../coords') as f:
        for line in f:
            line = line.strip().split()
            org,chromo,start,end,ori,hit_name = line
            org = org_mapping[org]
            try:
                chromo = chr_mapping[org][chromo]
            except:
                print(f'there are no anchors on {chromo}...')
                print(f'not putting element of {org_mapping[org]} on {chromo} starting at {start} ending at {end} in MCScanX files (will be neglected for analysis)')
                continue
            if ori == 'forward':
                strand = '+'
            elif ori == 'reverse':
                strand = '-'
            ele_name = f'{chromo}CUSTOMele{start}to{end}{strand}'
            coords.append((chromo,ele_name,int(start),int(end)))
            if chromo not in coords_per_chr:
                coords_per_chr[chromo] = []
            coords_per_chr[chromo].append((int(start),int(end),ele_name))
            
    for chromo in coords_per_chr:
        coords_per_chr[chromo].sort()

    filtered_coords_per_chr = {}
    to_del = set()
    for chromo,l in coords_per_chr.items():
        if len(l) == 1:
            start1,end1,ele_name1 = l[0]
            if chromo not in filtered_coords_per_chr:
                filtered_coords_per_chr[chromo] = [(start1,end1)]
        else:
            for enu,t in enumerate(l[:-1]):
                start1,end1,ele_name1 = t
                if (chromo,ele_name1,start1,end1) in to_del:
                    continue
                for start2,end2,ele_name2 in l[enu+1:]:
                    if end1 < start2:
                        if chromo in filtered_coords_per_chr:
                            filtered_coords_per_chr[chromo].append((start1,end1))
                        else:
                            filtered_coords_per_chr[chromo] = [(start1,end1)]
                        break
                    else:
                        to_del.add((chromo,ele_name2,start2,end2))
            if chromo not in filtered_coords_per_chr:
                lens = [x[1] - x[0] for x in l]
                idx = lens.index(max(lens))
                filtered_coords_per_chr[chromo] = l[idx:idx+1]
            if l[-2][1] < filtered_coords_per_chr[chromo][-1][0]:
                filtered_coords_per_chr[chromo].append((l[-1][0],l[-1][1]))
        filtered_coords_per_chr[chromo].sort()
    for ele in to_del:
        coords.remove(ele)
    
    min_max_per_chr = {}
    for chromo,l in filtered_coords_per_chr.items():
        min_max_per_chr[chromo] = (l[0][0]-margin,l[-1][1]+margin)

    elements,take = write_gff()
    write_homology(elements,take)



#!/usr/bin/env python3
import re
import json
import subprocess
import sys
import time
import os

def get_species_name(accession):
    """Return organism name for a given accession using NCBI datasets CLI."""
    try:
        result = subprocess.run(
            [os.path.expanduser("~/datasets"), "summary", "genome", "accession", accession],
            capture_output=True, text=True, check=True
        )
        data = json.loads(result.stdout)
        return data["reports"][0]["organism"]["organism_name"]
    except Exception as e:
        print(f"⚠️  Failed to fetch {accession}: {e}", file=sys.stderr)
        return None

def main():
    input_file = "tree"
    output_file = "tree_with_species.txt"

    with open(input_file) as f:
        tree_text = f.read()

    accessions = sorted(set(re.findall(r"(GC[FA]_\d+\.\d+)", tree_text)))
    print(f"🔍 Found {len(accessions)} accessions.")

    mapping = {}
    for acc in accessions:
        print(f"→ {acc}", end=": ")
        species = get_species_name(acc)
        if species:
            formatted = f"{species.replace(' ', '_')}[{acc}]"
            mapping[acc] = formatted
            print(formatted)
        else:
            mapping[acc] = acc
            print("⚠️  (kept accession)")
        time.sleep(0.2)

    def replace_func(m):
        acc = m.group(1)
        return mapping.get(acc, acc)

    new_tree = re.sub(r"(GC[FA]_\d+\.\d+)", replace_func, tree_text)

    with open(output_file, "w") as f:
        f.write(new_tree)

    print(f"\n✅ Done! Wrote output to '{output_file}'.")
    for k, v in list(mapping.items())[:5]:
        print(f"  {k} → {v}")

if __name__ == "__main__":
    main()




#! /usr/bin/env python3

from sys import argv
from subprocess import run
import pickle,os
from os.path import isfile


if os.path.exists('../anchors/'):
    anchors_path = '../anchors/'
elif os.path.exists('../utils/anchors/'):
    anchors_path = '../utils/anchors/'
else:
    raise FileNotFoundError("Anchors not found")

refs = []
with open('refs.txt','r') as f:
    for line in f:
        refs.append(line.strip())

target = argv[1]

weights = {}
max_w = 0
for ref in refs:
    aligned_len = 0
    with open(f'{anchors_path}/aligned/{ref}','rb') as f:
        anchor_map = pickle.load(f)
    for i,d in anchor_map.items():
        if target in d['matches']:
            if 'matches' in d['matches'][target]:
                if d['matches'][target]['meta']['multiple matches out of tolerance range'] == 1 or d['matches'][target]['meta']['matches have ambiguous matches (tolerance/chromosome out of range)'] == 1:
                    continue
                for j,bib_m in d['matches'][target]['matches'].items():
                    aligned_len += bib_m[f'hit coordinates in {target} candidate'][1] - bib_m[f'hit coordinates in {target} candidate'][0]
    weights[ref] = aligned_len
    if aligned_len > max_w:
        max_w = aligned_len

while max_w > 1e5:
    print(max_w)
    for ref in weights:
        if int(weights[ref]) == int(max_w):
            max_w = int(max_w/10)
        weights[ref] = int(weights[ref]/10)
        if weights[ref] < 1:
            weights[ref] = 1


with open('ref_weights.txt','w') as f:
    for ref,aligned_len in weights.items():
        f.write(f'{ref}\t{int(aligned_len)}\n')



#! /usr/bin/env python3


import sys
sys.path.append('../../utils/')
from get_mapping import get_mapping
import pickle
from pygenomeviz import GenomeViz
from subprocess import run
from pprint import pprint
import argparse
from os.path import isfile

def turn(start,end,segments):
    max_e = 0
    min_s = 1e15
    for segment in segments:
        max_e = max(segment[1],max_e)
        min_s = min(segment[0],min_s)
    len_line = max_e - min_s
    new_end = len_line - start
    new_start = new_end - (end - start)
    return new_start, new_end

def get_gff():

    gff = {}
    with open('MCScanX.gff') as f:
        for line in f:
            chromo,name,start,end = line.strip().split()
            gff[name] = (chromo,start,end)

    return gff

if __name__ == "__main__":

    parser= argparse.ArgumentParser(description='Draw MCScanX chains. The order maximise pariwise aligned length greedily.',formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--mcscanx_file', required=True, type=str,nargs=1, default='MCScanX.collinearity',help="MCScanX collinearity file")
    parser.add_argument('--threshold_chr', required=True, type=int,nargs=1, default=1000000,help="Threshold size for Chromosome/Scaffold/Contig to include in drawing")
    args = parser.parse_args()
    mcscanx_file = args.mcscanx_file[0]
    threshold_chr = args.threshold_chr[0]
    
    run(f'mkdir -p singles_out',shell=True)
    run(f'mkdir -p simple_maps',shell=True)
    run(f'mkdir -p orders',shell=True)

    run(f'python3 get_mc_blocks.py {mcscanx_file}',shell=True)
    run(f'python3 get_orders.py {threshold_chr}',shell=True)
   
    tr = []
    with open('orders_orgs') as f:
        for line in f:
            target,ref = line.strip().split('\t')
            tr.append((target,ref))


    org_mapping,chr_mapping = get_mapping()

    gff = get_gff()
    # save if using multiple times
    #with open('parsed_gff','wb') as f:
    #    pickle.dump(gff,f)
    #with open('parsed_gff','rb') as f:
    #    gff = pickle.load(f)

    drawing_order = {}
    for target,ref in tr:
        drawing_order[target] = []
        if not isfile(f'orders/{target}'):
            continue
        with open(f'orders/{target}') as f:
            for line in f:
                seq, ori = line.strip().split('\t')
                drawing_order[target].append((seq,ori))

    small_meta = {}
    for org in [x[0] for x in tr]:
        with open(f'../../utils/small_meta/{org}','rb') as f:
            small_meta[org] = pickle.load(f)
    with open(f'../../utils/small_meta/{tr[0][1]}','rb') as f:
        small_meta[tr[0][1]] = pickle.load(f)

    gv = GenomeViz()
    new_lengths = {}
    cur = 0
    target_ranges = []
    draw_ch = {}
    draw_ch[tr[0][1]] = {}
    new_lengths[tr[0][1]] = {}
    orig_lengths = []
    tmporder = []
    for enu,ch in enumerate(small_meta[tr[0][1]][0]):
        if enu == 0:
            if small_meta[tr[0][1]][1][enu] < int(threshold_chr):
                continue
            else:
                draw_ch[tr[0][1]][ch] = 1
                target_ranges.append((cur,cur+small_meta[tr[0][1]][1][enu]))
                new_lengths[tr[0][1]][ch] = (cur,cur+small_meta[tr[0][1]][1][enu])
                orig_lengths.append((0,small_meta[tr[0][1]][1][enu]))
                cur = cur+small_meta[tr[0][1]][1][enu]
                tmporder.append(ch)
        elif small_meta[tr[0][1]][1][enu] - small_meta[tr[0][1]][1][enu-1] < int(threshold_chr):
            continue
        else:
            draw_ch[tr[0][1]][ch] = 1
            target_ranges.append((cur,cur+small_meta[tr[0][1]][1][enu] - small_meta[tr[0][1]][1][enu-1]))
            new_lengths[tr[0][1]][ch] = (cur,cur+small_meta[tr[0][1]][1][enu] - small_meta[tr[0][1]][1][enu-1])
            orig_lengths.append((small_meta[tr[0][1]][1][enu-1] + 1,small_meta[tr[0][1]][1][enu]))
            cur = cur+small_meta[tr[0][1]][1][enu] - small_meta[tr[0][1]][1][enu-1] + 1
            tmporder.append(ch)
    track = gv.add_feature_track(tr[0][1],segments=target_ranges)

    for enu,segment in enumerate(track.segments):
            segment.add_sublabel('')#f'{tmporder[enu]}')
    track.set_segment_sep(symbol="//")

    for org in [x[0] for x in tr]:
        cur = 0
        target_ranges = []
        draw_ch[org] = {}
        new_lengths[org] = {}
        orig_lengths = []
        tmporder = []
        for seq,ori in drawing_order[org]:
            enu = small_meta[org][0].index(seq)
            if enu == 0:
                if small_meta[org][1][enu] < int(threshold_chr):
                    continue
                else:
                    draw_ch[org][seq] = 1
                    target_ranges.append((cur,cur+small_meta[org][1][enu]))
                    new_lengths[org][seq] = (cur,cur+small_meta[org][1][enu])
                    orig_lengths.append((0,small_meta[org][1][enu]))
                    cur = cur+small_meta[org][1][enu]
                    tmporder.append(seq)
            elif small_meta[org][1][enu] - small_meta[org][1][enu-1] < int(threshold_chr):
                continue
            else:
                draw_ch[org][seq] = 1
                target_ranges.append((cur,cur+small_meta[org][1][enu] - small_meta[org][1][enu-1]))
                new_lengths[org][seq] = (cur,cur+small_meta[org][1][enu] - small_meta[org][1][enu-1])
                orig_lengths.append((max(0,small_meta[org][1][enu-1] + 1),small_meta[org][1][enu]))
                cur = cur+small_meta[org][1][enu] - small_meta[org][1][enu-1] + 1
                tmporder.append(seq)
        track = gv.add_feature_track(org,segments=target_ranges)

        for enu,segment in enumerate(track.segments):
            segment.add_sublabel('')#f'{tmporder[enu]}')
        track.set_segment_sep(symbol="//")
   

    draw = 0


    with open('naive_blocks.pickle','rb') as f:
        blocks = pickle.load(f)

    homology = {}
    for b in blocks:
        t1,t2,rev = b
        first1,last1 = t1
        first2,last2 = t2
        c11,s11,e11 = gff[first1]
        c12,s12,e12 = gff[last1]
        c21,s21,e21 = gff[first2]
        c22,s22,e22 = gff[last2]
        s11 = int(s11)
        e11 = int(e11)
        s12 = int(s12)
        e12 = int(e12)
        s21 = int(s21)
        e21 = int(e21)
        s22 = int(s22)
        e22 = int(e22)
        org1 = org_mapping[first1.split('ele')[0].split('chr')[0]]
        org2 = org_mapping[first2.split('ele')[0].split('chr')[0]]
        chromo1 = chr_mapping[org1][c11]
        chromo2 = chr_mapping[org2][c21]
        start1 = s11
        end1 = e12
        start2 = s21
        end2 = e22
        if org1 not in homology:
            homology[org1] = {}
        if chromo1 not in homology[org1]:
            homology[org1][chromo1] = {}
        if start1 not in homology[org1][chromo1]:
            homology[org1][chromo1][start1] = {'matches': {}}
        homology[org1][chromo1][start1]['matches'][org2] = (chromo2,start2,rev,end1)

        if org2 not in homology:
            homology[org2] = {}
        if chromo2 not in homology[org2]:
            homology[org2][chromo2] = {}
        if start2 not in homology[org2][chromo2]:
            homology[org2][chromo2][start2] = {'matches': {}}
        homology[org2][chromo2][start2]['matches'][org1] = (chromo1,start1,rev,end2)
    
    for org1,org2 in tr:
        with open(f'simple_maps/{org1}','rb') as f:
            simple_map = pickle.load(f)
        for seq,ori in drawing_order[org1]:
            if seq not in homology[org1]:
                continue
            enu = small_meta[org1][0].index(seq)
            for start1,bib in homology[org1][seq].items():
                if org2 in bib['matches']:
                    chromo2,start2,rev,end1 = bib['matches'][org2]
                    if chromo2 not in draw_ch[org2]:
                        continue
                    end2 = homology[org2][chromo2][start2]['matches'][org1][3]
                    nstart1 = new_lengths[org1][seq][0] + start1
                    nend1 = new_lengths[org1][seq][0] + end1
                    nstart2 = new_lengths[org2][chromo2][0] + start2
                    nend2 = new_lengths[org2][chromo2][0] + end2
                    if rev == 1:
                        nstart2,nend2 = nend2,nstart2
                    if chromo2 == simple_map[seq][0]:
                        col = "blue"
                        inv_col = "lime"
                        al = 0.1
                    else:
                        col = "magenta"
                        inv_col = "orange"
                        al = 0.9
                    for track in gv.feature_tracks:
                        if track.name != org1:
                            continue
                        for segment in track.segments:
                            segstart = int(segment.start)
                            segend = int(segment.end)
                            if nstart1 < segstart or nend1 > segend:
                                continue
                            else:
                                seg1 = segment.name
                    for track in gv.feature_tracks:
                        if track.name != org2:
                            continue
                        for segment in track.segments:
                            segstart = int(segment.start)
                            segend = int(segment.end)
                            if nstart2 < segstart or nend2 > segend:
                                continue
                            else:
                                seg2 = segment.name

                    gv.add_link((org1,seg1,nstart1,nend1),(org2,seg2,nstart2,nend2), color=col, inverted_color=inv_col, curve=True, alpha = al)
                    draw += 1

fig = gv.plotfig()
fig.savefig(f'mcscx_blocks.svg', dpi=300)



#! /usr/bin/env python3

from subprocess import run
import os
import argparse
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

def get_genome(genomes_dir,org):
    seqs = SeqIO.parse(f'{genomes_dir}/{org}.fasta', "fasta")
    s = {}
    for seq in seqs:
        s[seq.id] = seq.seq
    return(s)

def translate_seq(seq,frame,orientation):
    # handle reverse orientation
    if orientation == 'reverse':
        seq = seq.reverse_complement()

    # translate based on frame
    if frame is None:
        # no frame - try all 3 and pick longest
        proteins = []
        for start_pos in [0,1,2]:
            prot = seq[start_pos:].translate(to_stop=False)
            proteins.append(prot)
        return max(proteins, key=len)
    else:
        # use specified frame
        frame_abs = abs(frame)
        start_pos = (frame_abs - 1) % 3
        return seq[start_pos:].translate(to_stop=False)

def deduplicate_overlapping_hits(hits, min_reciprocal_overlap=0.5, max_size_ratio=2.0):
    """
    Deduplicate overlapping hits that likely represent the same genomic element.

    Strategy:
    - Group hits by chromosome
    - Cluster overlapping hits based on reciprocal overlap and size ratio
    - Keep best representative (longest sequence) per cluster
    - Use union of coordinates for genomic range

    Args:
        hits: List of hit dictionaries with keys: chromo, start, end, seq, gene_id, etc.
        min_reciprocal_overlap: Minimum reciprocal overlap to consider merging (default 0.5)
        max_size_ratio: Maximum size ratio to consider merging (default 2.0)

    Returns:
        Deduplicated list of hits
    """
    from collections import defaultdict

    # Group hits by chromosome
    by_chrom = defaultdict(list)
    for hit in hits:
        by_chrom[hit['chromo']].append(hit)

    deduplicated = []

    for chromo, chrom_hits in by_chrom.items():
        # Sort by start position
        chrom_hits.sort(key=lambda h: h['start'])

        # Track which hits have been merged
        merged = [False] * len(chrom_hits)

        for i in range(len(chrom_hits)):
            if merged[i]:
                continue

            # Start a new cluster with hit i
            cluster = [i]

            # Find all hits that should be merged with hit i
            for j in range(i + 1, len(chrom_hits)):
                if merged[j]:
                    continue

                # Check if any hit in cluster overlaps with hit j
                should_merge = False
                for cluster_idx in cluster:
                    h1 = chrom_hits[cluster_idx]
                    h2 = chrom_hits[j]

                    # Calculate overlap
                    overlap_start = max(h1['start'], h2['start'])
                    overlap_end = min(h1['end'], h2['end'])
                    overlap_length = max(0, overlap_end - overlap_start)

                    if overlap_length > 0:
                        # Calculate reciprocal overlap
                        len1 = h1['end'] - h1['start']
                        len2 = h2['end'] - h2['start']
                        recip_overlap = overlap_length / min(len1, len2)

                        # Calculate size ratio
                        size_ratio = max(len1, len2) / min(len1, len2)

                        # Merge if reciprocal overlap is high AND size ratio is reasonable
                        if recip_overlap >= min_reciprocal_overlap and size_ratio <= max_size_ratio:
                            should_merge = True
                            break

                if should_merge:
                    cluster.append(j)
                    merged[j] = True

            # Process cluster: pick best representative
            cluster_hits = [chrom_hits[idx] for idx in cluster]

            # Score by sequence length (longer = better/more complete)
            # For hits without sequences, use genomic length
            def score_hit(h):
                if h['seq'] is not None and len(h['seq']) > 0:
                    return len(h['seq'])
                else:
                    return h['end'] - h['start']

            best_hit = max(cluster_hits, key=score_hit)

            # Create merged hit with:
            # - Union of coordinates (earliest start, latest end)
            # - Best hit's sequence
            # - Best hit's gene_id
            merged_hit = {
                'org': best_hit['org'],
                'chromo': best_hit['chromo'],
                'start': min(h['start'] for h in cluster_hits),
                'end': max(h['end'] for h in cluster_hits),
                'orientation': best_hit['orientation'],
                'gene_id': best_hit['gene_id'],
                'seq': best_hit['seq'],
                'frame': best_hit.get('frame')
            }

            deduplicated.append(merged_hit)
            merged[i] = True

    return deduplicated

def process_org(org,input_dir,genomes_dir,output_dir,mode,deduplicate_overlaps=False):

    temp_coords_file = f'{input_dir}/temp_coords_{org}'
    if not os.path.isfile(temp_coords_file):
        return(0)

    # parse temp_coords
    # Format: org chromo start end orientation gene_id [sequence]
    # sequence column can be empty for blastn (will extract from genome)
    hits = []
    with open(temp_coords_file,'r') as f:
        for line in f:
            if not line.strip():
                continue
            cols = line.strip().split('\t')

            # temp_coords must have exactly 7 columns
            if len(cols) != 7:
                raise ValueError(f"temp_coords line has {len(cols)} columns, expected 7 (org chromo start end orientation gene_id sequence). Line: {line.strip()}")

            hit = {
                'org': cols[0],
                'chromo': cols[1],
                'start': int(cols[2]),
                'end': int(cols[3]),
                'orientation': cols[4],
                'gene_id': cols[5]
            }
            # check if sequence is provided (column 6, can be empty for blastn)
            if len(cols[6]) > 0:
                # sequence provided - use it directly (tblastn case)
                hit['seq'] = cols[6]
                hit['frame'] = None
            else:
                # no sequence - will extract from genome (blastn case)
                hit['seq'] = None
                hit['frame'] = None
            hits.append(hit)

    if len(hits) == 0:
        return(0)

    # Deduplicate overlapping hits if requested
    if deduplicate_overlaps:
        original_count = len(hits)
        hits = deduplicate_overlapping_hits(hits)
        if len(hits) < original_count:
            print(f"  {org}: Deduplicated {original_count} hits to {len(hits)} ({original_count - len(hits)} merged)")

    # load genome
    genome = get_genome(genomes_dir,org)

    # extract sequences
    sequences = {}
    for hit in hits:
        chromo = hit['chromo']
        start = hit['start']
        end = hit['end']
        gene_id = hit['gene_id']
        orientation = hit['orientation']
        frame = hit.get('frame')
        provided_seq = hit.get('seq')

        # if sequence provided in temp_coords and protein mode, use it directly
        if provided_seq is not None and mode == 'protein':
            seq_obj = Seq(provided_seq)
            record = SeqRecord(seq_obj, id=gene_id, description="")
            sequences[gene_id] = (record,chromo,start,end,orientation)
            continue

        # otherwise extract from genome
        if chromo not in genome:
            continue

        nucl_seq = genome[chromo][start:end]
        if len(nucl_seq) == 0:
            continue

        if mode == 'protein':
            # no provided sequence - translate from genome
            try:
                prot_seq = translate_seq(nucl_seq,frame,orientation)
            except:
                continue
            record = SeqRecord(prot_seq, id=gene_id, description="")
            sequences[gene_id] = (record,chromo,start,end,orientation)
        else:
            # nucleotide mode - always use full genomic region
            if orientation == 'reverse':
                nucl_seq = nucl_seq.reverse_complement()
            record = SeqRecord(nucl_seq, id=gene_id, description="")
            sequences[gene_id] = (record,chromo,start,end,orientation)

    if len(sequences) == 0:
        return(0)

    # write gff
    run(f'mkdir -p {output_dir}/gff',shell=True)
    with open(f'{output_dir}/gff/{org}.gff','w') as f:
        for gene_id,(record,chromo,start,end,orientation) in sorted(sequences.items()):
            # Write proper GFF3 format: seqid source type start end score strand phase attributes
            # Use CDS as type so run_synthology_prot.py will parse it
            strand_symbol = '+' if orientation == 'forward' else '-'
            f.write(f'{chromo}\t.\tCDS\t{start}\t{end}\t.\t{strand_symbol}\t.\tID={gene_id}\n')

    # write sequences
    extension = 'faa' if mode == 'protein' else 'fna'
    seq_dir = 'proteins' if mode == 'protein' else 'nucleotides'
    run(f'mkdir -p {output_dir}/{seq_dir}',shell=True)
    records = [record for gene_id,(record,chromo,start,end,orientation) in sorted(sequences.items())]
    SeqIO.write(records, f'{output_dir}/{seq_dir}/{org}.{extension}', 'fasta')

    return(len(sequences))

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Convert BLAST results to synthology format',formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('input_dir', type=str, help="directory with temp_coords files")
    parser.add_argument('--genomes_dir', type=str, nargs='?', default='../../utils/genomes', help="directory with genome FASTA files")
    parser.add_argument('--output_dir', type=str, nargs='?', default='annotations_for_synthology_from_regions', help="output directory")
    parser.add_argument('--mode', type=str, nargs='?', default='protein', choices=['protein','nucleotide'], help="protein or nucleotide mode")
    parser.add_argument('--deduplicate_overlaps', action='store_true', help="merge overlapping hits that likely represent the same element (useful for multi-copy gene families)")

    args = parser.parse_args()

    input_dir = args.input_dir
    genomes_dir = args.genomes_dir
    output_dir = args.output_dir
    mode = args.mode
    deduplicate_overlaps = args.deduplicate_overlaps

    path = os.getcwd()

    orgs = []
    if os.path.isfile('orgs'):
        with open('orgs') as f:
            for line in f:
                org = line.strip()
                if os.path.isfile(f'{input_dir}/temp_coords_{org}'):
                    orgs.append(org)
    else:
        # fallback: list all temp_coords files
        if os.path.isdir(input_dir):
            for filename in os.listdir(input_dir):
                if filename.startswith('temp_coords_'):
                    org = filename.replace('temp_coords_','')
                    orgs.append(org)

    print(f'{len(orgs)} organisms with temp_coords files')

    if deduplicate_overlaps:
        print("Deduplication of overlapping hits is ENABLED")

    for org in orgs:
        n = process_org(org,input_dir,genomes_dir,output_dir,mode,deduplicate_overlaps)
        print(f'{org}: {n} sequences extracted')

    print('Done')



#! /usr/bin/env python3

from ete3 import Tree
from sys import argv

tree = Tree(argv[1], format=1)

# Set midpoint root
tree.set_outgroup(tree.get_midpoint_outgroup())

# Write rooted tree
tree.write(outfile=argv[2], format=1)




#! /usr/bin/env python3

"""
Generate plotting order for draw_verbose.py based on alignment coverage.

This script calculates alignment coverage between organisms and chromosomes,
then generates a plotting_order file with (org, chromo) pairs ordered by
alignment relationships. Uses a greedy approach: start with first organism,
add its chromosomes, then move to the organism with highest alignment coverage.

Usage:
    ./generate_plotting_order.py

Reads from:
    - orgs: list of organism IDs
    - anchors/aligned/{org}: aligned anchor data for each organism
    - small_meta/{org}: chromosome metadata for each organism

Writes to:
    - plotting_order: chromosome-level plotting order (org chromo pairs)
"""

import pickle
import os
from os.path import isfile

def get_alignment_coverage(org, orgs, maps_dir='anchors/aligned', meta_dir='small_meta'):
    """
    Calculate alignment coverage from one organism to all others.

    Returns dict mapping org -> total aligned length (as fraction of org's genome)
    """
    res = {}

    # Load genome sizes
    size = {}
    for o in orgs:
        meta_found = False
        meta_paths_tried = []
        for meta_path in [meta_dir, f'../../utils/{meta_dir}', f'../../../utils/{meta_dir}']:
            meta_file = f'{meta_path}/{o}'
            meta_paths_tried.append(meta_file)
            try:
                with open(meta_file, 'rb') as f:
                    seqids, seqlen = pickle.load(f)
                size[o] = seqlen[-1]
                meta_found = True
                break
            except FileNotFoundError:
                continue
        if not meta_found:
            print(f'Warning: metadata for {o} not found in: {", ".join(meta_paths_tried)}')
            continue

    # Load alignment map for this organism
    map_found = False
    map_paths_tried = []
    for maps_path in [maps_dir, f'../../utils/{maps_dir}', f'../../../utils/{maps_dir}', f'../../utils/anchors/aligned', f'../../../utils/anchors/aligned']:
        map_file = f'{maps_path}/{org}'
        map_paths_tried.append(map_file)
        try:
            with open(map_file, 'rb') as f:
                anchor_map = pickle.load(f)
            map_found = True
            break
        except FileNotFoundError:
            continue
    if not map_found:
        print(f'Warning: alignment data for {org} not found in: {", ".join(map_paths_tried)}')
        return res

    # Calculate aligned lengths
    aligned = {}
    scores = {}
    for o in orgs:
        if o == org:
            continue
        aligned[o] = 0
        scores[o] = []

    for i, d in anchor_map.items():
        for o, d2 in d['matches'].items():
            if o not in orgs:
                continue
            if d2['meta']['multiple matches out of tolerance range'] == 1 or d2['meta']['matches have ambiguous matches (tolerance/chromosome out of range)'] == 1:
                continue
            if 'matches' in d2:
                for j, bib_m in d2['matches'].items():
                    scores[o].append(bib_m['match score'])
                    aligned[o] += bib_m[f'hit coordinates in (own) {org} candidate'][1] - bib_m[f'hit coordinates in (own) {org} candidate'][0]
            if 'dups_matches' in d2 and 'syntenic' in d2['dups_matches']:
                js = d2['dups_matches']['syntenic']
                for j in js:
                    bib_m = d2['dups_matches'][j]
                    scores[o].append(bib_m['match score'])
                    aligned[o] += bib_m[f'hit coordinates in (own) {org} candidate'][1] - bib_m[f'hit coordinates in (own) {org} candidate'][0]

    for o in orgs:
        if o == org:
            continue
        if len(scores[o]) > 0 and org in size:
            res[o] = aligned[o] / size[org]

    return res


def get_chromosomes_for_org(org, maps_dir='anchors/aligned'):
    """
    Get list of chromosomes for an organism that have alignment data.

    Returns list of chromosome IDs.
    """
    anchor_map = None
    map_paths_tried = []
    for maps_path in [maps_dir, f'../../utils/{maps_dir}', f'../../../utils/{maps_dir}', f'../../utils/anchors/aligned', f'../../../utils/anchors/aligned']:
        map_file = f'{maps_path}/{org}'
        map_paths_tried.append(map_file)
        try:
            with open(map_file, 'rb') as f:
                anchor_map = pickle.load(f)
            break
        except FileNotFoundError:
            continue

    if anchor_map is None:
        print(f'Warning: alignment data for {org} not found in: {", ".join(map_paths_tried)}')
        return []

    chromosomes = set()
    for i, d in anchor_map.items():
        if 'chromosome' in d:
            chromosomes.add(d['chromosome'])

    return sorted(list(chromosomes))


if __name__ == "__main__":
    # Read organism list - try multiple locations
    orgs = []
    orgs_paths = ['orgs', '../../utils/orgs', '../../../utils/orgs']
    orgs_file_found = False

    for orgs_path in orgs_paths:
        if os.path.isfile(orgs_path):
            with open(orgs_path) as f:
                for line in f:
                    orgs.append(line.strip())
            orgs_file_found = True
            print(f'Loaded {len(orgs)} organisms from {orgs_path}')
            break

    # Fallback: auto-detect from pairwise_alignments_table
    if not orgs_file_found or len(orgs) == 0:
        print('Warning: orgs file not found, auto-detecting organisms from pairwise_alignments_table...')
        orgs_set = set()
        if os.path.isfile('pairwise_alignments_table'):
            with open('pairwise_alignments_table') as f:
                for line in f:
                    if line.strip():
                        cols = line.strip().split('\t')
                        if len(cols) >= 2:
                            orgs_set.add(cols[0])
                            orgs_set.add(cols[1])
        orgs = sorted(list(orgs_set))
        print(f'Auto-detected {len(orgs)} organisms from pairwise_alignments_table')

    if len(orgs) == 0:
        print('Error: no organisms found')
        exit(1)

    # Calculate pairwise alignment coverage between all organisms
    print('Calculating alignment coverage between organisms...')
    gres = {}
    for org in orgs:
        gres[org] = get_alignment_coverage(org, orgs)

    # Greedily order organisms by alignment coverage
    # Start with first organism, then pick organism with highest coverage
    orders = []
    last = orgs[0]
    while len(orders) != len(orgs) - 1:
        maxi = [0, '']
        for o in orgs:
            if o not in gres[last] or gres[last][o] == 0:
                continue
            if gres[last][o] > maxi[0] and o not in [xxx[1] for xxx in orders] and o != orgs[0]:
                maxi = [gres[last][o], o]
        if maxi[1] == '':
            # No more organisms with alignments, add remaining orgs
            for o in orgs:
                if o not in [xxx[0] for xxx in orders] and o not in [xxx[1] for xxx in orders] and o != orgs[0]:
                    maxi = [0, o]
                    break
        if maxi[1] != '':
            orders.append((last, maxi[1]))
            last = maxi[1]
        else:
            break

    # Create chromosome-level plotting order
    # For each organism in order, add all its chromosomes
    plotting_order = []
    orgs_in_order = [orgs[0]]
    for ref, target in orders:
        if target not in orgs_in_order:
            orgs_in_order.append(target)

    for org in orgs_in_order:
        chromosomes = get_chromosomes_for_org(org)
        for chromo in chromosomes:
            plotting_order.append((org, chromo))

    # Write plotting order to file
    with open('plotting_order', 'w') as f:
        for org, chromo in plotting_order:
            f.write(f'{org} {chromo}\n')

    print(f'Generated plotting_order with {len(plotting_order)} chromosome tracks')
    print(f'Organism order: {" -> ".join(orgs_in_order)}')



#! /usr/bin/env python3

import pickle
import sys
import os

def main():
    if len(sys.argv) < 2:
        print("Usage: python3 make_element_tables.py <output_dir>")
        sys.exit(1)

    outd = sys.argv[1]
    stats_file = f'{outd}/genetic_elements_stats/element_statistics.pickle'

    if not os.path.exists(stats_file):
        print(f"ERROR: {stats_file} not found")
        print("Run compute_element_stats.py first")
        sys.exit(1)

    print("Loading element statistics...")
    with open(stats_file, 'rb') as f:
        data = pickle.load(f)

    element_stats = data['element_stats']
    per_species_stats = data['per_species_stats']
    species_list = data['species_list']

    # Create output directory
    table_dir = f'{outd}/genetic_elements_stats/tables'
    os.makedirs(table_dir, exist_ok=True)

    # Table 1: Global summary
    global_table = f'{table_dir}/global_element_summary.tex'
    total_elements = sum(per_species_stats[sp]['total_elements'] for sp in species_list)
    num_singletons = len(element_stats['singletons'])
    num_single_copy = sum(len(g) for g in element_stats['single_copy_groups'])
    num_multi_copy = sum(len(g) for g in element_stats['multi_copy_groups'])
    num_self_dup = len(element_stats['within_species_duplicates'])

    with open(global_table, 'w') as f:
        f.write("\\documentclass[11pt,a4paper]{article}\n")
        f.write("\\usepackage[utf8]{inputenc}\n")
        f.write("\\usepackage{booktabs}\n")
        f.write("\\usepackage[margin=1in]{geometry}\n")
        f.write("\\begin{document}\n")
        f.write("\\begin{table}[h]\n")
        f.write("\\centering\n")
        f.write("\\begin{tabular}{lr}\n")
        f.write("\\toprule\n")
        f.write("Category & Count \\\\\n")
        f.write("\\midrule\n")
        f.write(f"Total elements & {total_elements} \\\\\n")
        f.write(f"Singletons & {num_singletons} \\\\\n")
        f.write(f"In single-copy groups & {num_single_copy} \\\\\n")
        f.write(f"In multi-copy groups & {num_multi_copy} \\\\\n")
        f.write(f"Within-species duplicates* & {num_self_dup} \\\\\n")
        f.write("\\midrule\n")
        f.write(f"Single-copy ortholog groups & {len(element_stats['single_copy_groups'])} \\\\\n")
        f.write(f"Multi-copy ortholog groups & {len(element_stats['multi_copy_groups'])} \\\\\n")
        f.write("\\bottomrule\n")
        f.write("\\end{tabular}\n")
        f.write("\\caption{Global element classification summary}\n")
        f.write("\\end{table}\n")
        f.write("\\end{document}\n")

    print(f"Saved global summary table: {global_table}")

    # Table 2: Per-species summary
    per_species_table = f'{table_dir}/per_species_element_summary.tex'
    with open(per_species_table, 'w') as f:
        f.write("\\documentclass[11pt,a4paper]{article}\n")
        f.write("\\usepackage[utf8]{inputenc}\n")
        f.write("\\usepackage{booktabs}\n")
        f.write("\\usepackage[margin=1in]{geometry}\n")
        f.write("\\begin{document}\n")
        f.write("\\begin{table}[h]\n")
        f.write("\\centering\n")
        f.write("\\small\n")
        f.write("\\begin{tabular}{lrrrrrr}\n")
        f.write("\\toprule\n")
        f.write("Species & Input & Filtered & Final & Singletons & Single-copy & Multi-copy \\\\\n")
        f.write("\\midrule\n")

        for sp in species_list:
            stats = per_species_stats[sp]
            sp_short = sp.replace('GCF_', '')
            total_input = stats['total_elements'] + stats.get('filtered', 0)
            filtered = stats.get('filtered', 0)
            f.write(f"{sp_short} & {total_input} & {filtered} & {stats['total_elements']} & {stats['singletons']} & ")
            f.write(f"{stats['in_single_copy_groups']} & {stats['in_multi_copy_groups']} \\\\\n")

        f.write("\\bottomrule\n")
        f.write("\\end{tabular}\n")
        f.write("\\caption{Per-species element classification. Input = elements in input data, Filtered = removed during pipeline, Final = in final orthology graph}\n")
        f.write("\\end{table}\n")
        f.write("\\end{document}\n")

    print(f"Saved per-species summary table: {per_species_table}")

    # Table 3: Top ortholog partners per species
    partner_table = f'{table_dir}/per_species_top_partners.tex'
    with open(partner_table, 'w') as f:
        f.write("\\documentclass[11pt,a4paper]{article}\n")
        f.write("\\usepackage[utf8]{inputenc}\n")
        f.write("\\usepackage{booktabs}\n")
        f.write("\\usepackage[margin=1in]{geometry}\n")
        f.write("\\begin{document}\n")
        f.write("\\begin{table}[h]\n")
        f.write("\\centering\n")
        f.write("\\small\n")
        f.write("\\begin{tabular}{llr}\n")
        f.write("\\toprule\n")
        f.write("Species & Top Partner & Edges \\\\\n")
        f.write("\\midrule\n")

        for sp in species_list:
            stats = per_species_stats[sp]
            sp_short = sp.replace('GCF_', '')
            if stats['species_connectivity']:
                top_partner, top_count = stats['species_connectivity'].most_common(1)[0]
                partner_short = top_partner.replace('GCF_', '')
                f.write(f"{sp_short} & {partner_short} & {top_count} \\\\\n")
            else:
                f.write(f"{sp_short} & -- & 0 \\\\\n")

        f.write("\\bottomrule\n")
        f.write("\\end{tabular}\n")
        f.write("\\caption{Top ortholog partner per species (by edge count)}\n")
        f.write("\\end{table}\n")
        f.write("\\end{document}\n")

    print(f"Saved top partners table: {partner_table}")

    print(f"\nAll tables saved to {table_dir}/")

if __name__ == '__main__':
    main()



#! /usr/bin/env python3


import pickle
import sys
from os.path import isfile
from get_mapping import get_mapping

if __name__ == "__main__":

    anchor_dir = sys.argv[1]
    orgs = []
    with open('orgs','r') as f:
        for line in f:
            orgs.append(line.strip())

    org_mapping,chr_mapping = get_mapping()

    out = open('pairwise_alignments_table','w')

    done = {}
    c_id = 1
    score1against2 = 'NA'
    score2against1 = 'NA'

    candidates = {}
    for org in orgs:
        with open(f'{anchor_dir}/candidates/{org}','rb') as f:
            candidates[org] = pickle.load(f)

    for org in orgs:
        if isfile(f'{anchor_dir}/aligned/{org}_with_syn_eval'):
            ad = f'{anchor_dir}/aligned/{org}_with_syn_eval'
        else:
            ad = f'{anchor_dir}/aligned/{org}'
        with open(ad,'rb') as f:
            bib = pickle.load(f)
        done[org] = 1
        for i,bib2 in bib.items():
            chromo1 = bib2['chromosome'] 
            for org2,mbib in bib2['matches'].items():
                if mbib['meta']['multiple matches out of tolerance range'] == 1 or mbib['meta']['matches have ambiguous matches (tolerance/chromosome out of range)'] == 1:
                    continue
                if org2 not in orgs or org2 in done:
                    continue
                if 'matches' in mbib:
                    for j,m in mbib['matches'].items():
                        chromo2 = candidates[org2][j]['chromosome']
                        start1,end1 = m[f'hit coordinates in (own) {org} candidate']
                        start1 += bib2['start']
                        end1 += bib2['start']
                        start2,end2 = m[f'hit coordinates in {org2} candidate']
                        start2 += candidates[org2][j]['start']
                        end2 += candidates[org2][j]['start']
                        score = m['match score']
                        ori = m['match is on other strand in other genome']
                        if ori == True:
                            ori = 'reverse'
                        else:
                            ori = 'forward'
                        out.write(f'{c_id}\t{org}\t{chromo1}\t{start1}\t{end1}\t{org2}\t{chromo2}\t{start2}\t{end2}\t{ori}\t{score}\t{score1against2}\t{score2against1}\n')
                        c_id += 1
                if 'dups_matches' in mbib and 'syntenic' in mbib['dups_matches']:
                    syn = mbib['dups_matches']['syntenic']
                    for j in syn:
                        m = mbib['dups_matches'][j]
                        chromo2 = candidates[org2][j]['chromosome']
                        start1,end1 = m[f'hit coordinates in (own) {org} candidate']
                        start1 += bib2['start']
                        end1 += bib2['start']
                        start2,end2 = m[f'hit coordinates in {org2} candidate']
                        start2 += candidates[org2][j]['start']
                        end2 += candidates[org2][j]['start']
                        score = m['match score']
                        ori = m['match is on other strand in other genome']
                        if ori == True:
                            ori = 'reverse'
                        else:
                            ori = 'forward'
                        out.write(f'{c_id}\t{org}\t{chromo1}\t{start1}\t{end1}\t{org2}\t{chromo2}\t{start2}\t{end2}\t{ori}\t{score}\t{score1against2}\t{score2against1}\n')
                        c_id += 1
    out.close()



#! /usr/bin/env python3

from sys import argv
from subprocess import run
import os
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

def get_genome(genomes_dir,org):
    seqs = SeqIO.parse(f'{genomes_dir}/{org}.fasta', "fasta")
    s = {}
    for seq in seqs:
        s[seq.id] = seq.seq
    return(s)

def extract_fastas(org,regions,genomes_dir,output_dir):

    genome = get_genome(genomes_dir,org)

    records = []
    for chromo,start,end in regions:
        if chromo not in genome:
            continue
        seq = genome[chromo][start:end]
        record = SeqRecord(seq, id=f"{org}_{chromo}_{start}_{end}_forward", description="")
        records.append(record)

    run(f'mkdir -p {output_dir}/{org}',shell=True)
    SeqIO.write(records, f'{output_dir}/{org}/all.fasta', 'fasta')

    return(len(records))

if __name__ == "__main__":

    # args: syntenic_regions_file genomes_dir output_dir
    if len(argv) < 2:
        print("Usage: python extract_syntenic_fastas.py syntenic_regions_succinct [genomes_dir] [output_dir]")
        exit(1)

    regions_file = argv[1]
    genomes_dir = argv[2] if len(argv) > 2 else '../../utils/genomes'
    output_dir = argv[3] if len(argv) > 3 else 'fastas'

    path = os.getcwd()

    # parse syntenic regions
    regions_by_org = {}
    with open(regions_file,'r') as f:
        for line in f:
            if line.startswith('org') or not line.strip():
                continue
            cols = line.strip().split('\t')
            if len(cols) < 4:
                continue
            org,chromo,start,end = cols[0],cols[1],int(cols[2]),int(cols[3])
            if org not in regions_by_org:
                regions_by_org[org] = []
            regions_by_org[org].append((chromo,start,end))

    # extract fastas
    for org in regions_by_org:
        regions = regions_by_org[org]
        n = extract_fastas(org,regions,genomes_dir,output_dir)
        print(f'{org}: {n} regions extracted')



#! /usr/bin/env python3

from subprocess import run
import sys
sys.path.append('../../utils/')
from get_mapping import get_mapping
import pickle
from sys import argv

if __name__ == "__main__":

    margin = int(argv[1])
    if len(argv) > 2:
        mode = '_' + argv[2]
    else:
        mode = ''

    org_mapping,chr_mapping = get_mapping()
    segments_genes = {}
    segments_multiplicons = {}
    genes_multiplicons = {}

    ### check which segments dont only contain CUSTOM elements and are therefore valid ###
    first = 1
    segments = {}
    with open(f'out_only_relevant_{margin}{mode}/list_elements.txt') as f:
        for line in f:
            if first == 1:
                first = 0
                continue
            # id  segment gene    position    orientation
            id,segment,gene,position,orientation = line.strip().split()
            if segment not in segments:
                segments[segment] = set()
            segments[segment].add(gene)
    
    valid_segments = set()
    for segment,genes in segments.items():
        n_customs = 0
        for g in genes:
            if 'CUSTOM' in g:
                n_customs += 1
        if n_customs < len(genes) / 2:
            valid_segments.add(segment)

    first = 1
    segments = {}
    with open(f'out_only_relevant_{margin}{mode}/list_elements.txt') as f:
        for line in f:
            if first == 1:
                first = 0
                continue
            # id  segment gene    position    orientation
            id,segment,gene,position,orientation = line.strip().split()
            if segment not in valid_segments:
                continue
            gene = gene+orientation
            if 'CUSTOM' in gene:
                if gene not in segments_genes:
                    segments_genes[gene] = set()
                if segment not in segments:
                    segments[segment] = set()
                segments_genes[gene].add((segment,position))
                segments[segment].add(gene)

    first = 1
    with open(f'out_only_relevant_{margin}{mode}/segments.txt') as f:
        for line in f:
            if first == 1:
                first = 0
                continue
            # id  multiplicon genome  list    first   last    order
            id,multiplicon,genome,list,first,last,order = line.strip().split()
            if id in segments:
                if multiplicon not in genes_multiplicons:
                    genes_multiplicons[multiplicon] = set()
                    segments_multiplicons[multiplicon] = set()
                segments_multiplicons[multiplicon].add(id)
                genes_multiplicons[multiplicon].update(segments[id])

    alloc = {}
    for m,s in genes_multiplicons.items():
        pos = {}
        for gene in s:
            for segment,position in segments_genes[gene]:
                if segment in segments_multiplicons[m]:
                    if gene in pos:
                        print(gene,pos[gene],position)
                        input()
                    pos[gene] = position
        for gene,p in pos.items():
            org1 = gene.split('chr')[0]
            for gene2,p2 in pos.items():
                org2 = gene2.split('chr')[0]
                if gene != gene2 and p == p2 and org1 != org2:
                    if org1 not in alloc:
                        alloc[org1] = {}
                    if org2 not in alloc:
                        alloc[org2] = {}
                    if gene not in alloc[org1]:
                        alloc[org1][gene] = {}
                    if gene2 not in alloc[org2]:
                        alloc[org2][gene2] = {}
                    if org2 not in alloc[org1][gene]:
                        alloc[org1][gene][org2] = set()
                    if org1 not in alloc[org2][gene2]:
                        alloc[org2][gene2][org1] = set()
                    alloc[org1][gene][org2].add(gene2)
                    alloc[org2][gene2][org1].add(gene)
    
                    
    run(f'rm -rf pairwise_orthologies_{margin}{mode} && mkdir -p pairwise_orthologies_{margin}{mode}',shell=True) 
    for org,bib in alloc.items():
        org = org_mapping[org]
        with open(f'pairwise_orthologies_{margin}{mode}/{org}','w') as f:
            for ele,bib2 in bib.items():
                f.write(f'------------------------ element: {ele} ------------------------\n')
                f.write(f'------------------------ aligns with ------------------------\n')
                for org2,s in bib2.items():
                    org2 = org_mapping[org2]
                    f.write(f'\t --- {org2} ---\n')
                    f.write(f'\t\t {s}\n')
    with open(f'alloc_pairwise_{margin}{mode}','wb') as f:
        pickle.dump(alloc,f)



#! /usr/bin/env python3

import pickle
import sys
import os
import networkx as nx
from collections import defaultdict, Counter

def main():
    if len(sys.argv) < 2:
        print("Usage: python3 compute_element_stats.py <output_dir>")
        sys.exit(1)

    outd = sys.argv[1]

    # Load the final union graph (no new edges version)
    union_file = f'{outd}/final_union_graph_no_new_edges.pickle'
    if not os.path.exists(union_file):
        print(f"ERROR: {union_file} not found")
        sys.exit(1)

    print("Loading final union graph...")
    with open(union_file, 'rb') as f:
        data = pickle.load(f)
        if len(data) == 3:
            graph, gene_species, metadata = data
            print(f"Threshold: {metadata.get('threshold', 'N/A')}")
        else:
            graph, gene_species = data

    print(f"Loaded graph with {graph.number_of_nodes()} nodes and {graph.number_of_edges()} edges")
    print(f"Gene-species mapping: {len(gene_species)} genes")

    # Identify genes not in graph (filtered during pipeline)
    genes_in_graph = set(graph.nodes())
    genes_not_in_graph = set(gene_species.keys()) - genes_in_graph
    print(f"Genes filtered out during pipeline: {len(genes_not_in_graph)}")

    # Load filtering report if available
    filter_report_file = f'{outd}/genetic_elements_stats/filtering_report.pickle'
    filtering_reasons = {}
    if os.path.exists(filter_report_file):
        print(f"Loading filtering analysis from filtering_report.pickle...")
        with open(filter_report_file, 'rb') as f:
            filter_data = pickle.load(f)
            gene_status_all = filter_data.get('gene_status', {})

            # Categorize filtered genes by reason
            for gene in genes_not_in_graph:
                status = gene_status_all.get(gene, {})
                if not status.get('in_chains', False):
                    filtering_reasons[gene] = 'not_in_chains'
                elif not status.get('homology', False):
                    filtering_reasons[gene] = 'no_homology'
                elif not status.get('in_pairwise', False):
                    filtering_reasons[gene] = 'filtered_at_pairwise'
                elif not status.get('in_aligned', False):
                    filtering_reasons[gene] = 'filtered_at_alignment'
                elif not status.get('in_union_new', False):
                    filtering_reasons[gene] = 'filtered_at_union'
                else:
                    filtering_reasons[gene] = 'filtered_at_final_cograph'

        # Count by reason
        reason_counts = Counter(filtering_reasons.values())
        print(f"  Filtering reasons:")
        for reason, count in reason_counts.most_common():
            print(f"    - {reason}: {count} genes")
    else:
        print(f"  Run trace_filtered_genes.py first for detailed filtering analysis")

    # Get all species
    species_list = sorted(set(gene_species.values()))
    print(f"Species: {len(species_list)}")

    # Find connected components
    components = list(nx.connected_components(graph))
    print(f"Connected components: {len(components)}")

    # Classify elements
    element_stats = {
        'singletons': [],
        'single_copy_groups': [],
        'multi_copy_groups': [],
        'within_species_duplicates': [],
        'filtered_genes': list(genes_not_in_graph)
    }

    per_species_stats = {sp: {
        'total_elements': 0,  # Only genes in final graph
        'filtered': 0,  # Genes removed during pipeline
        'filtered_reasons': Counter(),  # Breakdown of filtering reasons
        'singletons': 0,
        'in_single_copy_groups': 0,
        'in_multi_copy_groups': 0,
        'self_duplicated': 0,
        'species_connectivity': Counter(),  # Count edges to other species
        'num_species_connected': Counter()  # Count how many elements connect to N species
    } for sp in species_list}

    # Count total elements per species (only those in graph)
    for gene, sp in gene_species.items():
        if gene in genes_in_graph:
            per_species_stats[sp]['total_elements'] += 1
        else:
            per_species_stats[sp]['filtered'] += 1
            # Track filtering reason if available
            if gene in filtering_reasons:
                per_species_stats[sp]['filtered_reasons'][filtering_reasons[gene]] += 1

    # Analyze each component
    for comp in components:
        comp_genes = list(comp)
        comp_size = len(comp_genes)

        # Get species composition
        comp_species_counts = Counter([gene_species[g] for g in comp_genes])
        num_species = len(comp_species_counts)

        # Classify component type
        is_singleton = comp_size == 1
        is_single_copy = all(count == 1 for count in comp_species_counts.values())
        is_multi_copy = any(count > 1 for count in comp_species_counts.values())

        if is_singleton:
            gene = comp_genes[0]
            sp = gene_species[gene]
            element_stats['singletons'].append(gene)
            per_species_stats[sp]['singletons'] += 1
        elif is_single_copy:
            element_stats['single_copy_groups'].append(comp_genes)
            for gene in comp_genes:
                sp = gene_species[gene]
                per_species_stats[sp]['in_single_copy_groups'] += 1
        else:  # multi-copy
            element_stats['multi_copy_groups'].append(comp_genes)
            for gene in comp_genes:
                sp = gene_species[gene]
                per_species_stats[sp]['in_multi_copy_groups'] += 1

                # Check if this gene itself is duplicated within species
                same_species_in_comp = [g for g in comp_genes if gene_species[g] == sp]
                if len(same_species_in_comp) > 1:
                    per_species_stats[sp]['self_duplicated'] += 1
                    if gene not in element_stats['within_species_duplicates']:
                        element_stats['within_species_duplicates'].append(gene)

        # Analyze connectivity for each gene in component
        for gene in comp_genes:
            sp = gene_species[gene]

            # Count connections to other species
            species_connected = set()
            for neighbor in graph.neighbors(gene):
                neighbor_sp = gene_species[neighbor]
                if neighbor_sp != sp:
                    per_species_stats[sp]['species_connectivity'][neighbor_sp] += 1
                    species_connected.add(neighbor_sp)

            # Track how many species this element connects to
            num_connected_species = len(species_connected)
            per_species_stats[sp]['num_species_connected'][num_connected_species] += 1

    # Print global statistics
    total_in_graph = graph.number_of_nodes()
    print("\n" + "="*60)
    print("GLOBAL ELEMENT STATISTICS")
    print("="*60)
    print(f"Total elements in input: {len(gene_species)}")
    print(f"Filtered during pipeline: {len(genes_not_in_graph)} ({100*len(genes_not_in_graph)/len(gene_species):.1f}%)")
    print(f"In final graph: {total_in_graph} ({100*total_in_graph/len(gene_species):.1f}%)")
    print(f"\nClassification of {total_in_graph} elements in final graph:")
    print(f"  Singletons: {len(element_stats['singletons'])} ({100*len(element_stats['singletons'])/total_in_graph:.1f}%)")
    print(f"  In single-copy groups: {sum(len(g) for g in element_stats['single_copy_groups'])} ({100*sum(len(g) for g in element_stats['single_copy_groups'])/total_in_graph:.1f}%)")
    print(f"  In multi-copy groups: {sum(len(g) for g in element_stats['multi_copy_groups'])} ({100*sum(len(g) for g in element_stats['multi_copy_groups'])/total_in_graph:.1f}%)")
    print(f"  Within-species duplicates*: {len(element_stats['within_species_duplicates'])}")

    print(f"\nOrthology groups:")
    print(f"  Single-copy groups: {len(element_stats['single_copy_groups'])}")
    print(f"  Multi-copy groups: {len(element_stats['multi_copy_groups'])}")

    # Print per-species statistics
    print("\n" + "="*60)
    print("PER-SPECIES ELEMENT STATISTICS")
    print("="*60)

    for sp in species_list:
        stats = per_species_stats[sp]
        total_with_filtered = stats['total_elements'] + stats['filtered']
        print(f"\n{sp}:")
        print(f"  Total in input: {total_with_filtered}")
        if stats['filtered'] > 0:
            print(f"  Filtered: {stats['filtered']}")
            if stats['filtered_reasons']:
                for reason, count in stats['filtered_reasons'].most_common():
                    print(f"    - {reason}: {count}")
        print(f"  In final graph: {stats['total_elements']}")
        if stats['total_elements'] > 0:
            print(f"    Singletons: {stats['singletons']} ({100*stats['singletons']/stats['total_elements']:.1f}%)")
            print(f"    In single-copy groups: {stats['in_single_copy_groups']} ({100*stats['in_single_copy_groups']/stats['total_elements']:.1f}%)")
            print(f"    In multi-copy groups: {stats['in_multi_copy_groups']} ({100*stats['in_multi_copy_groups']/stats['total_elements']:.1f}%)")
            if stats['self_duplicated'] > 0:
                print(f"      - Self-duplicated*: {stats['self_duplicated']}")

        # Top species connections
        if stats['species_connectivity']:
            top_partners = stats['species_connectivity'].most_common(3)
            print(f"  Top ortholog partners:")
            for partner_sp, count in top_partners:
                print(f"    - {partner_sp}: {count} edges")

    # Save results for plotting/tables
    results = {
        'element_stats': element_stats,
        'per_species_stats': per_species_stats,
        'species_list': species_list,
        'graph': graph,
        'gene_species': gene_species
    }

    stats_dir = f'{outd}/genetic_elements_stats'
    os.makedirs(stats_dir, exist_ok=True)
    output_file = f'{stats_dir}/element_statistics.pickle'
    with open(output_file, 'wb') as f:
        pickle.dump(results, f)
    print(f"\nSaved element statistics to {output_file}")

if __name__ == '__main__':
    main()



#! /usr/bin/env python3

"""
Draw verbose synteny visualizations with pygenomeviz.

This script creates detailed synteny plots showing genomic regions of interest
with alignment links between organisms. It reads coordinates from a coords file,
pairwise alignment data, and determines chromosome orientations automatically.

Usage:
    ./draw_verbose.py MARGIN [REF_ORG REF_CHROMO REF_START] [OPTIONS]

    # Basic usage with margin
    ./draw_verbose.py 50000

    # With reference element
    ./draw_verbose.py 50000 GCF_016746395.2 NC_052520.2 80300

    # With custom paths
    ./draw_verbose.py 50000 --coords-file ../coords --alignments-file pairwise_alignments_table

    # Auto-generate plotting order
    ./draw_verbose.py 50000 --auto-order

    # Use GFF files instead of coords
    ./draw_verbose.py 50000 --gff-dir ~/test_out_dir/out/stable/synthology/halos/gff

    # Draw all elements from GFF (not just connected component)
    ./draw_verbose.py 50000 --gff-dir ~/test_out_dir/out/stable/synthology/halos/gff --draw-all-elements

Arguments:
    MARGIN: Integer margin around elements (required)
    REF_ORG, REF_CHROMO, REF_START: Reference element to focus on (optional)

Options:
    --coords-file: Path to coords file (default: ../coords)
    --alignments-file: Path to pairwise alignments table (default: pairwise_alignments_table)
    --output-dir: Output directory for images (default: images_draw_verbose)
    --auto-order: Auto-generate plotting_order file before drawing

Reads from:
    - coords file: genomic coordinates of elements of interest
    - pairwise_alignments_table: alignment data between organisms
    - plotting_order (optional): order of tracks in figure
    - try_to_draw_all (optional): flag to draw all chromosomes

Writes to:
    - images_draw_verbose/{margin}/{ref}_*.svg: synteny figure
    - tables_drawing/{org}.tsv: gene feature tables
"""

from sys import argv
import argparse
from matplotlib.pyplot import cm
from pygenomeviz import GenomeViz
from random import randint
import numpy as np
from subprocess import run
from pprint import pprint
import pickle
from matplotlib.lines import Line2D
from matplotlib.patches import Patch
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
from os.path import isfile

### IMPORTANT NOTES $$$

# Configuration constants
SEGMENT_GAP_THRESHOLD = 500000  # Split track into segments if gap exceeds this (bp)
ELEMENT_LABEL_MARGIN_TOP = 0.8
ELEMENT_LABEL_MARGIN_BOTTOM = 2.5

# 1. since the chromsome/contig lengths is not explicitely known here, the drawing can be longer than a chromosome/contig

def parse_gff_directory(gff_dir):
    import os
    from pathlib import Path

    orgs = []
    coords = {}
    to_ori = set()

    gff_files = sorted(Path(gff_dir).glob('*.gff'))
    if not gff_files:
        gff_files = sorted(Path(gff_dir).glob('*.gff3'))

    for gff_file in gff_files:
        # Extract organism ID from filename (e.g., GCF_000001215.4.gff -> GCF_000001215.4)
        org = gff_file.stem
        orgs.append(org)
        coords[org] = {}

        with open(gff_file) as f:
            for line in f:
                if line.startswith('#'):
                    continue

                fields = line.strip().split('\t')
                if len(fields) < 9:
                    continue

                chromo = fields[0]
                feature_type = fields[2]
                start = fields[3]
                end = fields[4]
                strand_symbol = fields[6]
                attributes = fields[8]

                # Only process CDS features
                if feature_type != 'CDS':
                    continue

                # Convert strand to orientation
                if strand_symbol == '+':
                    strand = 1
                    ori_gene = 'forward'
                elif strand_symbol == '-':
                    strand = -1
                    ori_gene = 'reverse'
                else:
                    continue

                # Extract gene name from attributes
                gene_name = None
                for attr in attributes.split(';'):
                    if attr.startswith('gene='):
                        gene_name = attr.split('=')[1]
                        break
                    elif attr.startswith('Name='):
                        gene_name = attr.split('=')[1]
                        break

                if not gene_name:
                    # Fall back to protein_id
                    for attr in attributes.split(';'):
                        if attr.startswith('protein_id='):
                            gene_name = attr.split('=')[1]
                            break

                if not gene_name:
                    gene_name = f"{chromo}_{start}_{end}"

                # Add to coords structure
                if chromo not in coords[org]:
                    coords[org][chromo] = []
                    to_ori.add((org, chromo))

                coords[org][chromo].append((int(start), int(end), strand, gene_name))

    return orgs, coords, to_ori


def turn(start,end,segments):
    max_e = 0
    min_s = 1e15
    for segment in segments:
        max_e = max(segment[1],max_e)
        min_s = min(segment[0],min_s)
    len_line = max_e - min_s
    new_end = len_line - start
    new_start = new_end - (end - start)
    return new_start, new_end

def overlaps(ranges):
    """Merge overlapping ranges and combine their color sets and strand info."""
    ranges = sorted(ranges)
    it = iter(ranges)
    try:
        curr_start, curr_stop, curr_color, curr_strands = next(it)
    except StopIteration:
        return
    for start, stop, color, strands in it:
        if curr_start <= start <= curr_stop:
            curr_stop = max(curr_stop, stop)
            curr_color.update(color)
            curr_strands.update(strands)
        else:
            yield curr_start, curr_stop, curr_color, curr_strands
            curr_start, curr_stop, curr_color, curr_strands = start, stop, color, strands
    yield curr_start, curr_stop, curr_color, curr_strands

# Parse command-line arguments
parser = argparse.ArgumentParser(
    description='Draw verbose synteny visualizations with pygenomeviz',
    formatter_class=argparse.RawDescriptionHelpFormatter,
    epilog="""
Examples:
  %(prog)s 50000
  %(prog)s 50000 GCF_016746395.2 NC_052520.2 80300
  %(prog)s 50000 --auto-order --coords-file ../coords
    """
)
parser.add_argument('margin', type=int, help='Margin around elements (bp)')
parser.add_argument('ref_org', nargs='?', default=None, help='Reference organism')
parser.add_argument('ref_chromo', nargs='?', default=None, help='Reference chromosome')
parser.add_argument('ref_start', nargs='?', default=None, help='Reference start position')
parser.add_argument('--coords-file', default='../coords', help='Path to coords file (default: ../coords)')
parser.add_argument('--gff-dir', default=None, help='Path to directory containing GFF files (use instead of coords-file)')
parser.add_argument('--alignments-file', default='pairwise_alignments_table', help='Path to pairwise alignments table (default: pairwise_alignments_table)')
parser.add_argument('--output-dir', default='images_draw_verbose', help='Output directory for images (default: images_draw_verbose)')
parser.add_argument('--auto-order', action='store_true', help='Auto-generate plotting_order file before drawing')
parser.add_argument('--draw-all-elements', action='store_true', help='Draw all elements instead of filtering by connectivity to reference')
parser.add_argument('--no-anchors', action='store_true', help='Skip drawing anchor rectangles (only draw alignment curves). Default: draw anchors.')

args = parser.parse_args()

# Auto-generate plotting order if requested
if args.auto_order:
    print('Auto-generating plotting_order...')
    result = run('./generate_plotting_order.py', shell=True)
    if result.returncode != 0:
        print('Warning: generate_plotting_order.py failed, continuing anyway')

margin = args.margin

# Set reference element if provided
if args.ref_org and args.ref_chromo and args.ref_start:
    ref = (args.ref_org, args.ref_chromo, args.ref_start)
else:
    ref = 0

# Parse coords from either GFF directory or coords file
orgs = []
coords = {}
to_ori = set()
ori = {}

if args.gff_dir:
    import os
    gff_dir_expanded = os.path.expanduser(args.gff_dir)
    print(f'Parsing GFF files from {gff_dir_expanded}...')
    orgs, coords, to_ori = parse_gff_directory(gff_dir_expanded)
    print(f'Found {len(orgs)} organisms with {sum(len(coords[org]) for org in orgs)} chromosomes')
    # Set default reference if not specified
    if ref == 0 and orgs:
        first_org = orgs[0]
        first_chromo = list(coords[first_org].keys())[0]
        first_start = coords[first_org][first_chromo][0][0]
        ref = (first_org, first_chromo, str(first_start))
        print(f'Using default reference: {ref}')
else:
    # Parse coords file (original format)
    with open(args.coords_file) as f:
        for line in f:
            line = line.strip().split()
            if len(line) == 5:
                org,chromo,start,end,ori_gene = line
                hit_name = f'{org}_{chromo}_{start}_{end}'
            elif len(line) == 6:
                org,chromo,start,end,ori_gene,hit_name = line
            else:
                continue

            if (chromo,org,start) == ref:
                ori[ref] = ori_gene
            if org not in coords:
                coords[org] = {}
                orgs.append(org)
            if chromo not in coords[org]:
                coords[org][chromo] = []
                to_ori.add((org,chromo))
            if ori_gene == 'forward':
                strand = 1
            else:
                strand = -1
            if ref == 0:
                ref = (org,chromo,start)
            coords[org][chromo].append((int(start),int(end),strand,hit_name))

run(f'mkdir -p tables_drawing',shell=True)
tables = {}
for org in orgs:
    tables[org] = open(f'tables_drawing/{org}.tsv','w')
    tables[org].write(f'species\tchromosome\tstart\tend\tstrand\tlabel\toriginal protein name (in ncbi annotation/proteome)\n')

tot_lens = {}
no_anchors = {}
for org in orgs:
    tot_lens[org] = 0
    no_anchors[org] = 0

bib = {}

no_colors = 0
with open(args.alignments_file) as f:
    for line in f:
        if line[0] == '#':
            continue
        no_colors += 1

colors = cm.rainbow(np.linspace(0, 1, no_colors))
used = {}
orientation = {}
alignments = {}

pairs = {}

with open(args.alignments_file) as f:
    for line in f:
        if line[0] == '#':
            continue
        no,org1,chromo1,start1,end1,org2,chromo2,start2,end2,ori_alignment,score,snd1,snd2 = line.split()
        if org1 not in coords or org2 not in coords or chromo1 not in coords[org1] or chromo2 not in coords[org2]:
            continue
        start1 = int(start1)
        end1 = int(end1)
        start2 = int(start2)
        end2 = int(end2)
        if start1 > end1:
            start1,end1 = end1,start1
        if start2 > end2:
            start2,end2 = end2,start2
        relevant1 = 0
        rel1 = []
        relevant2 = 0
        rel2 = []
        for t in coords[org1][chromo1]:
            start,end,strand,hit_name = t
            if start1 > start - margin and end1 < end + margin:
                relevant1 = 1
                rel1.append((org1,chromo1,t[0]))
        for t in coords[org2][chromo2]:
            start,end,strand,hit_name = t
            if start2 > start - margin and end2 < end + margin:
                relevant2 = 1
                rel2.append((org2,chromo2,t[0]))
        if relevant1 == 0 or relevant2 == 0:
            continue
        for ele1 in rel1:
            if ele1 not in pairs:
                pairs[ele1] = set()
            for ele2 in rel2:
                if ele2 not in pairs:
                    pairs[ele2] = set()
                pairs[ele1].add(ele2)
                pairs[ele2].add(ele1)

        if (org1,chromo1,start1,end1) not in alignments:
            alignments[(org1,chromo1,start1,end1)] = []
        alignments[(org1,chromo1,start1,end1)].append([org2,chromo2,start2,end2,ori_alignment])
        if (org2,chromo2,start2,end2) not in alignments:
            alignments[(org2,chromo2,start2,end2)] = []
        alignments[(org2,chromo2,start2,end2)].append([org1,chromo1,start1,end1,ori_alignment])
        if org1 not in bib:
            bib[org1] = {}
            orientation[org1] = {}
        if chromo1 not in bib[org1]:
            bib[org1][chromo1] = []
            orientation[org1][chromo1] = {}
        if org2 not in bib:
            bib[org2] = {}
            orientation[org2] = {}
        if chromo2 not in bib[org2]:
            bib[org2][chromo2] = []
            orientation[org2][chromo2] = {}
        if org1 not in orientation[org2][chromo2]:
            orientation[org2][chromo2][org1] = {}
        if chromo1 not in orientation[org2][chromo2][org1]:
            orientation[org2][chromo2][org1][chromo1] = []
        if org2 not in orientation[org1][chromo1]:
            orientation[org1][chromo1][org2] = {}
        if chromo2 not in orientation[org1][chromo1][org2]:
            orientation[org1][chromo1][org2][chromo2] = []
        ri = randint(0,no_colors-1)
        while ri in used:
            ri = randint(0,no_colors-1)
        used[ri] = 1
        color = ri
        bib[org1][chromo1].append((start1,end1,set([color])))
        bib[org2][chromo2].append((start2,end2,set([color])))
        orientation[org1][chromo1][org2][chromo2].append(ori_alignment)
        orientation[org2][chromo2][org1][chromo1].append(ori_alignment)

# Find connected components using BFS
def get_connected_component(start_ele, pairs_graph, visited):
    """BFS to find one connected component starting from start_ele."""
    component = set()
    component.add(start_ele)
    new_eles = pairs_graph[start_ele] if start_ele in pairs_graph else set()
    visited.add(start_ele)

    while len(new_eles) > 0:
        ne = set()
        for cur_ele in new_eles:
            if cur_ele in visited:
                continue
            visited.add(cur_ele)
            component.add(cur_ele)
            if cur_ele in pairs_graph:
                for ele2 in pairs_graph[cur_ele]:
                    if ele2 not in visited:
                        ne.add(ele2)
        new_eles = ne

    return component

# Determine which elements to draw
if args.draw_all_elements:
    # Find all connected components
    print('Finding all connected components (--draw-all-elements flag set)')
    all_elements = set(pairs.keys())
    visited = set()
    components = []

    for start_ele in all_elements:
        if start_ele not in visited:
            component = get_connected_component(start_ele, pairs, visited)
            components.append(component)

    print(f'Found {len(components)} connected components')
    for i, comp in enumerate(components):
        chromo_counts = {}
        for org, chromo, start in comp:
            key = f'{org}:{chromo}'
            chromo_counts[key] = chromo_counts.get(key, 0) + 1
        print(f'  Component {i+1}: {len(comp)} elements across {len(chromo_counts)} chromosomes')
        print(f'    Chromosomes: {dict(list(chromo_counts.items())[:5])}')

    # Will draw each component as a separate image
    if len(components) == 0:
        print('No elements with alignments found')
        exit(0)
else:
    # Filter by connectivity to reference element (default behavior)
    # Start from reference, expand to all elements connected through alignment pairs
    if (ref[0],ref[1],int(ref[2])) not in pairs or len(pairs[(ref[0],ref[1],int(ref[2]))]) == 0:
        print('exiting because focal element does not have alignments')
        exit(0)
    else:
        visited = set()
        component = get_connected_component((ref[0],ref[1],int(ref[2])), pairs, visited)
        components = [component]

# Save original bib before component loop (will be filtered per component)
import copy
original_bib = copy.deepcopy(bib)

# Process each connected component separately
for comp_idx, to_draw in enumerate(components):
    if args.draw_all_elements:
            print(f'\n=== Processing Component {comp_idx+1}/{len(components)} ({len(to_draw)} elements) ===')

    # Restore original bib for this component
    bib = copy.deepcopy(original_bib)

    new_bib = {}
    tdbib = {}
    for org,chromo,start in to_draw:
        if org not in tdbib:
            tdbib[org] = {}
        if chromo not in tdbib[org]:
            tdbib[org][chromo] = []
        for startele,end,strand,hit_name in coords[org][chromo]:
            if start == startele:
                tdbib[org][chromo].append((start,end,strand))

    for org,x in bib.items():
        if org not in tdbib:
            continue
        for chromo,l in x.items():
            if chromo not in tdbib[org]:
                continue

            # When using GFF files: check if anchor has alignments to component chromosomes
            # instead of proximity to elements (elements are sparse in GFF mode)
            if args.gff_dir:
                for ele_start, ele_end, color_set in l:
                    # Check if this anchor has any alignment to chromosomes in this component
                    has_component_alignment = False
                    if (org, chromo, ele_start, ele_end) in alignments:
                        for ali_org, ali_chromo, ali_start, ali_end, orient in alignments[(org, chromo, ele_start, ele_end)]:
                            if ali_org in tdbib and ali_chromo in tdbib[ali_org]:
                                has_component_alignment = True
                                break

                    if has_component_alignment:
                        if org not in new_bib:
                            new_bib[org] = {}
                        if chromo not in new_bib[org]:
                            new_bib[org][chromo] = []
                        new_bib[org][chromo].append((ele_start, ele_end, color_set))
            else:
                # Original behavior: only keep anchors near component elements
                for ele in l:
                    for gene in tdbib[org][chromo]:
                        if ele[0] >= gene[0]-margin and ele[1] <= gene[1]+margin:
                            if org not in new_bib:
                                new_bib[org] = {}
                            if chromo not in new_bib[org]:
                                new_bib[org][chromo] = []
                            new_bib[org][chromo].append(ele)

    bib = new_bib

    # Debug: show what's in bib after filtering
    if args.draw_all_elements:
        print(f'After filtering, bib contains {sum(len(v) for v in bib.values())} chromosome entries')
        for org in bib:
            print(f'  {org}: {list(bib[org].keys())}')

    # Determine chromosome orientations per connected component
    # Goal: minimize inversions across tracks in the figure
    # Pick first element in component as reference for this component
    comp_ref_ele = list(to_draw)[0]
    comp_ref = (comp_ref_ele[0], comp_ref_ele[1])  # (org, chromo)

    if comp_ref not in ori:
        ori[comp_ref] = 'forward'

    # Orient other chromosomes in this component relative to component reference
    for org,x in orientation.items():
        for chromo,y in x.items():
            if (org,chromo) in ori:
                continue
            # Check if this chromosome is in current component
            in_component = any((org, chromo, start) in to_draw for start in [t[2] for t in to_draw if t[0]==org and t[1]==chromo])
            if not in_component:
                continue

            if org not in orientation[comp_ref[0]][comp_ref[1]]:
                continue
            if chromo not in orientation[comp_ref[0]][comp_ref[1]][org]:
                continue
            oris = orientation[comp_ref[0]][comp_ref[1]][org][chromo]
            # Orient based on majority of alignment orientations to reference
            if oris.count('forward') > oris.count('reverse'):
                ori[(org,chromo)] = 'forward'
            else:
                ori[(org,chromo)] = 'reverse'

    # Orient remaining chromosomes based on already-oriented chromosomes
    # Propagate orientations through the alignment graph
    for org,x in orientation.items():
        for chromo,y in x.items():
            if (org,chromo) in ori:
                continue
            if org not in orientation[ref[0]][ref[1]]:
                continue
            if chromo not in orientation[ref[0]][ref[1]][org]:
                continue
            oris = orientation[ref[0]][ref[1]][org][chromo]
            # Orient based on majority of alignment orientations to reference
            if oris.count('forward') > oris.count('reverse'):
                ori[(org,chromo)] = 'forward'
            else:
                ori[(org,chromo)] = 'reverse'


    # Orient remaining chromosomes based on already-oriented chromosomes
    # Propagate orientations through the alignment graph
    for org,x in orientation.items():
        for chromo,y in x.items():
            if (org,chromo) in ori:
                continue
            for org2,z in y.items():
                for chromo2,oris in z.items():
                    if (org2,chromo2) not in ori:
                        continue
                    if org2 not in orientation[org][chromo]:
                        continue
                    if chromo2 not in orientation[org][chromo][org2]:
                        continue
                    oris = orientation[org][chromo][org2][chromo2]
                    # Determine orientation relative to already-oriented chromosome
                    if oris.count('forward') > oris.count('reverse'):
                        if ori[(org2,chromo2)] == 'forward':
                            ori[(org,chromo)] = 'forward'
                        else:
                            ori[(org,chromo)] = 'reverse'
                    else:
                        if ori[(org2,chromo2)] == 'forward':
                            ori[(org,chromo)] = 'reverse'
                        else:
                            ori[(org,chromo)] = 'forward'

    first = 1
    if isfile('try_to_draw_all'):
        while len(ori) != len(to_ori):
            if first == 1:
                for org,chromo in to_ori:
                    if (org,chromo) not in ori:
                        ori[(org,chromo)] = 'forward'
                        first = 0
                        break
            else:
                first = 1
                for org,x in orientation.items():
                    for chromo,y in x.items():
                        if (org,chromo) in ori:
                            continue
                        for org2,z in y.items():
                            for chromo2,oris in z.items():
                                if (org2,chromo2) not in ori:
                                    continue
                                if org2 not in orientation[org][chromo]:
                                    continue
                                if chromo2 not in orientation[org][chromo][org2]:
                                    continue
                                oris = orientation[org][chromo][org2][chromo2]
                                if oris.count('forward') > oris.count('reverse'):
                                    if ori[(org2,chromo2)] == 'forward':
                                        ori[(org,chromo)] = 'forward'
                                    else:
                                        ori[(org,chromo)] = 'reverse'
                                else:
                                    if ori[(org2,chromo2)] == 'forward':
                                        ori[(org,chromo)] = 'reverse'
                                    else:
                                        ori[(org,chromo)] = 'forward'

    for org,x in bib.items():
        for chromo,l in x.items():
            new_l = []
            for start,end,color_set in l:
                strands = set()
                alis = alignments[(org,chromo,start,end)]
                for ali in alis:
                    org2,chromo2,start2,end2,orient = ali
                    if orient == '*':
                        strands.add('reverse')
                        continue
                    if (org,chromo) not in ori:
                        continue
                    if (org2,chromo2) not in ori:
                        continue
                    ori1 = ori[(org,chromo)]
                    ori2 = ori[(org2,chromo2)]
                    if (ori1 == ori2 and ali[-1] == 'forward') or (ori1 != ori2 and ali[-1] == 'reverse'):
                        strands.add('forward')
                    else:
                        if ali[-1] == 'forward':
                            if ori1 == 'forward':
                                strands.add('forward')
                            else:
                                strands.add('reverse')
                        else:
                            strands.add('forward')
                            for enu,t in enumerate(alignments[(org2,chromo2,start2,end2)]):
                                a,b,c,d,e = t
                                if a == org and b == chromo and c == start and d == end:
                                    alignments[(org2,chromo2,start2,end2)][enu][-1] = '*'
                        
                new_l.append((start,end,color_set,strands))
            bib[org][chromo] = new_l

    # Merge overlapping alignment fragments and assign consistent colors
    # Fragments that overlap get the same color in the final figure
    color_sets = {}
    members = {}
    counter = 1

    for org,x in bib.items():
        for chromo,l in x.items():
            l = list(overlaps(l))
            for start,end,color_set,strands in l:
                # Find which color groups are already taken by overlapping fragments
                colors_taken = set()
                for color in color_set:
                    if color in members:
                        colors_taken.add(members[color])
                # Assign new color group or merge with existing groups
                if len(colors_taken) == 0:
                    # New color group
                    color_sets[counter] = color_set
                    for color in color_set:
                        members[color] = counter
                    counter += 1
                elif len(colors_taken) == 1:
                    # Merge into existing color group
                    taken = list(colors_taken)[0]
                    color_sets[taken].update(color_set)
                    for color in color_set:
                        members[color] = taken
                else:
                    # Multiple color groups need to be merged
                    color_sets[counter] = set()
                    for color in colors_taken:
                        color_sets[counter].update(color_sets[color])
                        for color2 in color_sets[color]:
                            members[color2] = counter
                        del color_sets[color]
                    color_sets[counter].update(color_set)
                    for color in color_set:
                        members[color] = counter
                    counter += 1
            bib[org][chromo] = l

    new_colors = iter(cm.rainbow(np.linspace(0, 1, len(color_sets))))
    colors = {}

    for org,x in bib.items():
        for chromo,l in x.items():
            new_l = []
            for start,end,color_set,strands in l:
                cont = 0
                for coord in coords[org][chromo]:
                    if (start >= coord[0] and start <= coord[1]) or (end >= coord[0] and end <= coord[1]) or (start <= coord[0] and end >= coord[1]) or (start >= coord[0] and end <= coord[1]):
                        cont = 1
                        break
                if cont == 1:
                    continue
                if len(set([members[color] for color in color_set])) != 1:
                    input('ERROR collecting all overlapping fragments and their colors')
                else:
                    color_no = list(set([members[color] for color in color_set]))[0]
                if color_no not in colors:
                    colors[color_no] = next(new_colors)
                color = colors[color_no]
                new_l.append((start,end,color,strands))
            bib[org][chromo] = new_l

    color_counter = 1
    used = {}
    gv = GenomeViz()
    shifts = {}
    genome_sizes = {}

    org_chromo_order = []
    if isfile('plotting_order'):
        with open('plotting_order') as f:
            for line in f:
                org,chromo = line.strip().split()
                org_chromo_order.append((org,chromo))
    else:
        for org,d in bib.items():
            for chromo in d:
                org_chromo_order.append((org,chromo))

    new_order = []
    for org,chromo in org_chromo_order:
        if org in bib and chromo in bib[org]:
            new_order.append((org,chromo))
    org_chromo_order = new_order

    # Assign default orientation 'forward' to any chromosomes that still don't have one
    for org, chromo in org_chromo_order:
        if (org, chromo) not in ori:
            ori[(org, chromo)] = 'forward'

    # Debug: show final plotting order
    if args.draw_all_elements:
        print(f'Final plotting order has {len(org_chromo_order)} tracks')
        unoriented_count = sum(1 for org, chromo in org_chromo_order if ori.get((org, chromo)) == 'forward')
        print(f'  ({unoriented_count} with default forward orientation)')

    for org,chromo in org_chromo_order:
        l = bib[org][chromo]
        if (org,chromo) not in ori or len(l) == 0:
            continue
        # first find elements and their size to make figures and get shift
        start = min(y[0] for y in l)
        end = max(y[1] for y in l)
        # i take anchors to actually check which genes are in the cluster
        if org not in coords or chromo not in coords[org]:
            continue
        start_min_ele = 1e15
        end_max_ele = 0
        for start_l,end_l,strand,hit_name in coords[org][chromo]:
            start_l,end_l = int(start_l),int(end_l)
            if (end_l < start and start-end_l > margin) or (start_l > end and start_l - end > margin):
                continue
            if start_l < start_min_ele:
                start_min_ele = start_l
            if end_l > end_max_ele:
                end_max_ele = end_l
        if start_min_ele == 1e15 and end_max_ele == 0:
            continue
        start = start_min_ele - margin
        local_margin_up = margin
        local_margin_down = margin
        if start < 0:
            local_margin_up = start_min_ele
            start = 0
        end = end_max_ele + margin
        if ori[(org,chromo)] == 'reverse':
            local_margin_up,local_margin_down = local_margin_down,local_margin_up
        shift = start
        genome_size = end - start
        name = org+' '+chromo + ' ' + ori[(org,chromo)]
        name = name.replace('_',' ')
        shifts[name] = shift

        gene_drawings = []
        for start_l,end_l,strand,hit_name in coords[org][chromo]:
            start_l,end_l = int(start_l),int(end_l)
            if end_l < start or start_l > end:
                continue
            start_l -= shift
            end_l -= shift
            if ori[(org,chromo)] == 'reverse':
                start_l,end_l = turn(start_l,end_l,[(0,genome_size)])
                if strand == -1:
                    strand = 1
                else:
                    strand = -1
            if start_l > genome_size or end_l > genome_size or start_l < 0 or end_l < 0:
                continue
            gene_drawings.append((max(0,start_l-local_margin_up),min(end_l+local_margin_down,genome_size)))

        gene_drawings = sorted(gene_drawings)
        target_ranges = []
        i = 1
        s1 = gene_drawings[0][0]
        e1 = gene_drawings[0][1]
        while i < len(gene_drawings):
            s2 = gene_drawings[i][0]
            e2 = gene_drawings[i][1]
            if s2 - e1 > SEGMENT_GAP_THRESHOLD:
                target_ranges.append((s1,e1))
                s1 = s2
            e1 = e2
            i+=1
        if len(target_ranges) == 0:
            target_ranges = (gene_drawings[0][0],gene_drawings[-1][1])
        else:
            if gene_drawings[-1][1] >  target_ranges[-1][1]:
                if target_ranges[-1][0] != s1:
                    target_ranges.append((s1,gene_drawings[-1][1]))
                else:
                    target_ranges.append((gene_drawings[-1][0],gene_drawings[-1][1]))

        track = gv.add_feature_track(name,segments=target_ranges)
        seg_labs = []
        genome_sizes[name] = []
        for segment in track.segments:
            if ori[(org,chromo)] == 'reverse':
                seg_labs.append((segment.start,segment.end))
            else:
                segment.add_sublabel(f'{segment.start+shift}-{segment.end+shift}')
            genome_sizes[name].append((int(segment.start),int(segment.end)))
            tot_lens[org] += int(segment.end) - int(segment.start)
        if len(seg_labs) > 0:
            # For reverse tracks, convert visualization coords back to genomic coords
            seg_labels = []
            seg_labs = sorted(seg_labs)
            for s,e in seg_labs:
                # Convert from visualization space back to genomic space
                # turn() did: new_end = genome_size - start; new_start = new_end - (end-start)
                # So reverse: genomic_start = shift + (genome_size - viz_end)
                #            genomic_end = shift + (genome_size - viz_start)
                genomic_start = shift + (genome_size - e)
                genomic_end = shift + (genome_size - s)
                # Show in decreasing order for reverse strand
                seg_labels.append(f'{int(genomic_end)} - {int(genomic_start)}')
            for i,s in enumerate(track.segments):
                s.add_sublabel(seg_labels[i])
        track.set_segment_sep(symbol="//")

        for start_l,end_l,strand,hit_name in coords[org][chromo]:
            orig_start = start_l
            orig_end = end_l
            start_l,end_l = int(start_l),int(end_l)
            if end_l < start or start_l > end:
                continue
            start_l -= shift
            end_l -= shift
            if strand == -1:
                vpos = 'bottom'
                ymargin = ELEMENT_LABEL_MARGIN_BOTTOM
            else:
                vpos = 'top'
                ymargin = ELEMENT_LABEL_MARGIN_TOP
            if ori[(org,chromo)] == 'reverse':
                start_l,end_l = turn(start_l,end_l,genome_sizes[name])
                if strand == -1:
                    strand = 1
                    vpos = 'top'
                    ymargin = ELEMENT_LABEL_MARGIN_TOP
                else:
                    strand = -1
                    vpos = 'bottom'
                    ymargin = ELEMENT_LABEL_MARGIN_BOTTOM
            
            label = hit_name
            facecolor = 'black'
            for segment in track.segments:
                try:
                    segment.add_feature(start_l,end_l,facecolor='black',label = label, text_kws=dict(color=facecolor,vpos=vpos,ymargin=ymargin,hpos='center'),strand = strand,plotstyle='arrow')
                    tables[org].write(f'{org}\t{chromo}\t{orig_start}\t{orig_end}\t{strand}\t{label}\t{hit_name}\n')
                except:
                    pass

        # Draw anchors (colored rectangles showing pairwise alignments)
        if not args.no_anchors:
            for start_l,end_l,color,strands in l:
                if start_l > end_l:
                    start_l,end_l = end_l,start_l
                cont = 0
                for coord in coords[org][chromo]:
                    if (start_l >= coord[0] and start_l <= coord[1]) or (end_l >= coord[0] and end_l <= coord[1]) or (start_l <= coord[0] and end_l >= coord[1]) or (start_l >= coord[0] and end_l <= coord[1]):
                        cont = 1
                        break
                if cont == 1:
                    continue
                start_l = start_l - shift
                end_l = end_l - shift
                if ori[(org,chromo)] == 'reverse':
                    start_l,end_l = turn(start_l,end_l,genome_sizes[name])
                if 'reverse' in strands:
                    strand_l = -1
                else:
                    strand_l = 1
                if tuple(color) not in used:
                    used[tuple(color)] = color_counter
                    color_counter += 1
                label = ''
                strand_l = 1
                added = 0
                for segment in track.segments:
                    try:
                        segment.add_feature(start_l,end_l,facecolor=color,strand = strand_l,plotstyle='arrow',label=label)
                        added = 1
                        no_anchors[org] += 1
                    except:
                        pass
                if added == 0:
                    print(f'CANNOT ADD anchor of {org} {chromo} {start_l} {end_l} {strands}...probably out of range: genome_size = {genome_size}')

    done = {}
    for t1,l in alignments.items():
        org,chromo,start1,end1 = t1
        if (org,chromo) not in ori:
            continue
        orig_start1 = start1
        orig_end1 = end1
        orig_start1 = int(orig_start1)
        orig_end1 = int(orig_end1)
        cont = 0
        for coord in coords[org][chromo]:
            if (start1 >= coord[0] and start1 <= coord[1]) or (end1 >= coord[0] and end1 <= coord[1]) or (start1 <= coord[0] and end1 >= coord[1]) or (start1 >= coord[0] and end1 <= coord[1]):
                cont = 1
                break
        if cont == 1:
            continue
        name1 = org+' '+chromo+' '+ori[(org,chromo)]
        name1 = name1.replace('_',' ')
        if name1 not in shifts:
            continue
        shift1 = shifts[name1]
        start1 -= shift1
        end1 -= shift1
        if ori[(org,chromo)] == 'reverse':
            start1,end1 = turn(start1,end1,genome_sizes[name1])
        for org2,chromo2,start2,end2,ori_alignment in l:
            if (org2,chromo2) not in ori:
                continue
            orig_start2 = start2
            orig_end2 = end2
            orig_start2 = int(orig_start2)
            orig_end2 = int(orig_end2)
            if ((org,chromo,orig_start1,orig_end1),(org2,chromo2,orig_start2,orig_end2)) in done or ((org2,chromo2,orig_start2,orig_end2),(org,chromo,orig_start1,orig_end1)) in done:
                continue
            else:
                done[((org,chromo,orig_start1,orig_end1),(org2,chromo2,orig_start2,orig_end2))] = 1
                done[((org2,chromo2,orig_start2,orig_end2),(org,chromo,orig_start1,orig_end1))] = 1
            cont = 0
            for coord in coords[org2][chromo2]:
                if (start2 >= coord[0] and start2 <= coord[1]) or (end2 >= coord[0] and end2 <= coord[1]) or (start2 <= coord[0] and end2 >= coord[1]) or (start2 >= coord[0] and end2 <= coord[1]):
                    cont = 1 
                    break
            if cont == 1:
                continue
            name2 = org2+' '+chromo2+ ' ' + ori[(org2,chromo2)]
            name2 = name2.replace('_',' ')
            if name2 not in shifts:
                continue
            shift2 = shifts[name2]
            start2 -= shift2
            end2 -= shift2
            if ori_alignment == 'reverse' and ori[(org,chromo)] == ori[(org2,chromo2)] or ori_alignment == 'forward' and ori[(org,chromo)] != ori[(org2,chromo2)]:
                start2,end2 = end2,start2
            if ori[(org2,chromo2)] == 'reverse':
                start2,end2 = turn(start2,end2,genome_sizes[name2])

            try:
                for track in gv.feature_tracks:
                    if track.name != name1:
                        continue
                    for segment in track.segments:
                        segstart = int(segment.start)
                        segend = int(segment.end)
                        if start1 < segstart or end1 > segend:
                            continue
                        else:
                            seg1 = segment.name
                for track in gv.feature_tracks:
                    if track.name != name2:
                        continue
                    for segment in track.segments:
                        segstart = int(segment.start)
                        segend = int(segment.end)
                        if start2 < segstart or end2 > segend:
                            continue
                        else:
                            seg2 = segment.name
                gv.add_link((name1,seg1,start1,end1),(name2,seg2,start2,end2), color="skyblue", inverted_color="lime", curve=True)
            except:
                print(f'link FROM {name1} {start1} {end1} TO {name2} {start2} {end2} not adjacent')

    run(f'mkdir -p {args.output_dir} && mkdir -p {args.output_dir}/{margin}',shell=True)
    fig = gv.plotfig()
    if args.draw_all_elements:
        # Find a representative element for naming
        first_ele = list(to_draw)[0]
        output_filename = f'{args.output_dir}/{margin}/component_{comp_idx+1}_of_{len(components)}_ref_{first_ele[0]}_{first_ele[1]}_{first_ele[2]}.svg'
    else:
        output_filename = f'{args.output_dir}/{margin}/{ref}_verbose_with_alignments_with_prot_labels.svg'
    fig.savefig(output_filename)
    print(f'Saved figure to {output_filename}')
    plt.close(fig)

    for org in orgs:
        tables[org].close()



#! /usr/bin/env python3

import pickle
import sys
sys.path.append('../../utils/')
from get_mapping import get_mapping
from subprocess import run
from pprint import pprint

def merge(c1,c2):
    c = c1 + c2
    return(list(set(c)))


if __name__ == "__main__":

    margin = int(sys.argv[1])

    org_mapping,chr_mapping = get_mapping()

    alloc = {}

    skip = {}
    lines = 0
    tot_eles = 0
    no_customs = 0

    with open(f'MCScanX_plus_coords_only_relevant_{margin}.collinearity','r') as f:
        for line in f:
            if 'Alignment' in line:
                if lines != 0 and no_customs < tot_eles / 2:#valid == 0:
                    for l in lines:
                        skip[l] = 1
                lines = []
                #valid = 0
                no_customs = 0
                tot_eles = 0
            else:
                tot_eles += 1
                if line.count('CUSTOM') != 2:
                    #valid = 1
                    no_customs += 1
                else:
                    lines.append(line)
    if no_customs < tot_eles / 2:#valid == 0:
        for l in lines:
            skip[l] = 1

    with open(f'MCScanX_plus_coords_only_relevant_{margin}.collinearity','r') as f:
        for line in f:
            if 'CUSTOM' in line and line not in skip:
                line = line.split(':')[1].split()
                first = line[0]
                org1 = first.split('chr')[0]
                if org1 not in alloc:
                    alloc[org1] = {}
                if first not in alloc[org1]:
                    alloc[org1][first] = {}
                second = line[1]
                org2 = second.split('chr')[0]
                if org2 not in alloc:
                    alloc[org2] = {}
                if second not in alloc[org2]:
                    alloc[org2][second] = {}
                if org1 not in alloc[org2][second]:
                    alloc[org2][second][org1] = set()
                if org2 not in alloc[org1][first]:
                    alloc[org1][first][org2] = set()
                alloc[org1][first][org2].add(second)
                alloc[org2][second][org1].add(first)
    
    run(f'rm -rf pairwise_orthologies_{margin} && mkdir -p pairwise_orthologies_{margin}',shell=True) 
    for org,bib in alloc.items():
        org = org_mapping[org]
        with open(f'pairwise_orthologies_{margin}/{org}','w') as f:
            for ele,bib2 in bib.items():
                f.write(f'------------------------ element: {ele} ------------------------\n')
                f.write(f'------------------------ aligns with ------------------------\n')
                for org2,s in bib2.items():
                    org2 = org_mapping[org2]
                    f.write(f'\t --- {org2} ---\n')
                    f.write(f'\t\t {s}\n')

    with open(f'alloc_pairwise_{margin}','wb') as f:
        pickle.dump(alloc,f)



#! /usr/bin/env python3

import os
dir_path = os.path.dirname(os.path.realpath(__file__))

def get_mapping():
    org_mapping = {}
    chr_mapping = {}
    if os.path.exists(dir_path+'/../utils/mapping'):
        ff = dir_path+'/../utils/mapping'
    elif os.path.exists(dir_path+'/../../utils/mapping'):
        ff = dir_path+'/../../utils/mapping'
    elif os.path.exists('mapping'):
        ff = dir_path+'/mapping'
    else:
        print(f'no mapping file found')
    with open(ff) as f:
        org = False
        for line in f:
            if line[0] == '#':
                if 'species' in line:
                    org = True
                continue
            if org:
                org_mapping[line.strip().split()[0]] = line.strip().split()[1]
                org_mapping[line.strip().split()[1]] = line.strip().split()[0]
                org = False
                which_org = line.strip().split()[0]
                chr_mapping[which_org] = {}
                chr_mapping[org_mapping[which_org]] = {}
                continue
            else:
                chr_mapping[which_org][line.strip().split()[0]] = line.strip().split()[1]
                chr_mapping[which_org][line.strip().split()[1]] = line.strip().split()[0]
                chr_mapping[org_mapping[which_org]][line.strip().split()[0]] = line.strip().split()[1]
                chr_mapping[org_mapping[which_org]][line.strip().split()[1]] = line.strip().split()[0]
    return(org_mapping,chr_mapping)



#! /usr/bin/env python3

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
from sys import argv
from subprocess import run
from pprint import pprint

def write_scaffolds(genome,scaffolds):

    mapping = {}
    
    new_genome = {}
    with open(scaffolds) as f:
        for line in f:
            if len(line.strip()) == 0:
                continue
            if '>' in line:
                new_genome[line.strip()] = []
                last = line.strip()
                place = 0
            elif len(line.split()) > 0:
                new_genome[last].append((line.strip()[:-1].strip(),line.strip()[-1].strip()))
                mapping[line.strip()[:-1].strip()] = (last,place,line.strip()[-1].strip())
                place += 1

    seq_lists = {}

    for new_scaffold,contigs in new_genome.items():
        seq_lists[new_scaffold] = ['' for i in range(len(contigs))]

    done = {}

    for seq in SeqIO.parse(genome,'fasta'):
        if seq.id in mapping:
            scaff,place,ori = mapping[seq.id]
            if ori == '+' or ori == '0':
                seq_lists[scaff][place] = str(seq.seq)
            elif ori == '-' or ori == '1':
                seq_lists[scaff][place] = str(seq.reverse_complement().seq)
            done[seq.id] = 1

    records = []
    join_str = "N"*100
    for new_scaffold,contigs in seq_lists.items():
        new_seq = join_str.join(contigs)
        records.append(SeqRecord(Seq(new_seq),id=f"{new_scaffold[1:]}",description=""))
    c = 0
    for seq in SeqIO.parse(genome,'fasta'):
        if seq.id not in done:
            c += 1
            records.append(seq)
    SeqIO.write(records, f'fastas/{outname}', "fasta")

    return(0)

if __name__ == "__main__":
    genome = argv[1].strip()
    scaffolds = argv[2].strip()
    outname = argv[3]
    run('mkdir -p fastas',shell=True)
    write_scaffolds(genome,scaffolds)



#! /usr/bin/env python3

import pickle
from sys import argv

borders = []
first = 1
with open(f'{argv[1]}','r') as f:
    for line in f:
        if line[0] == '#' and 'Alignment' not in line:
            continue
        elif 'Alignment' in line:
            if first == 1:
                first = 0
                matches = []
                maxi1 = [0,'']
                mini1 = [1e16,'']
                maxi2 = [0,'']
                mini2 = [1e16,'']
                continue
            else:
                if (int(matches[0][0].split('to')[1]) > int(matches[-1][0].split('to')[1])) or (int(matches[0][1].split('to')[1]) > int(matches[-1][1].split('to')[1])):
                    rev = 1
                else:
                    rev = 0
                first1 = mini1[1]
                last1 = maxi1[1]
                first2 = mini2[1]
                last2 = maxi2[1]
                borders.append(((first1,last1),(first2,last2),rev))
                matches = []
                maxi1 = [0,'']
                mini1 = [1e16,'']
                maxi2 = [0,'']
                mini2 = [1e16,'']
        else:
            line = line.split(':')[1].split()
            first = line[0].strip()
            second = line[1].strip()
            matches.append((first,second))
            start1 = int(first.split('ele')[1].split('to')[0])
            end1 = int(first.split('to')[1])
            start2 = int(second.split('ele')[1].split('to')[0])
            end2 = int(second.split('to')[1])
            if start1 < mini1[0]:
                mini1 = [start1,first]
            if end1 > maxi1[0]:
                maxi1 = [end1,first]
            if start2 < mini2[0]:
                mini2 = [start2,second]
            if end2 > maxi2[0]:
                maxi2 = [end2,second]

with open('naive_blocks.pickle','wb') as f:
    pickle.dump(borders,f)



#! /usr/bin/env python3

import pickle
import os
from pprint import pprint
from bisect import bisect_left
from copy import deepcopy

def overlaps(x,y):
    if x[0] >= y[0] and x[0] <= y[1]:
        return True
    if x[1] >= y[0] and x[1] <= y[1]:
        return True
    if y[0] >= x[0] and y[0] <= x[1]:
        return True
    if y[1] >= x[0] and y[1] <= x[1]:
        return True
    return False

if __name__ == "__main__":

    orgs = []
    with open('orgs') as f:
        for line in f:
            orgs.append(line.strip())

    for org in orgs:
        print(f'======= {org} =======')

        with open('../anchors/candidates' + f'/{org}','rb') as f:
            bib = pickle.load(f)

        print(len(bib))

        ses = sorted([(i,i+x['end']-x['start']) for i,x in bib.items()])
        starts = [x[0] for x in ses]
        ends = [x[1] for x in ses]

        # Check if dups file exists for this organism
        dups_file = f'dups/{org}'
        if os.path.exists(dups_file):
            with open(dups_file,'rb') as f:
                dups = pickle.load(f)

            to_del = []
            for i,d in dups.items():
                end = i + d['end']-d['start']

                end_bigger_start_of_dup = bisect_left(ends,i)
                start_smaller_end_of_dup = bisect_left(starts,end)

                new_entry = {}
                new_entry['i'] = i
                new_entry['bib'] = deepcopy(d)

                cur_start = i
                cur_end = end
                for idx in range(max(0,end_bigger_start_of_dup-3),min(len(ends),start_smaller_end_of_dup+4)):
                    if overlaps((starts[idx],ends[idx]),(cur_start,cur_end)):
                        if cur_end >= starts[idx] and cur_end - starts[idx] < (ends[idx] - starts[idx])/10:
                            new_entry['bib']['end'] -= (cur_end - starts[idx] + 100)
                            cur_end = cur_end - (cur_end-starts[idx] + 100)
                        elif cur_start <= ends[idx] and ends[idx] - cur_start < (ends[idx] - starts[idx])/10:
                            new_entry['i'] = cur_start + (ends[idx] - cur_start + 100)
                            new_entry['bib']['start'] += (ends[idx] - cur_start + 100)
                            cur_start = cur_start + (ends[idx] - cur_start + 100)
                        to_del.append(starts[idx])
                # some cases at chromosome starts/ends
                if cur_end - cur_start < 300 or cur_end < cur_start or cur_end < 0 or new_entry['bib']['end'] < 0 or new_entry['bib']['start'] < 0 or new_entry['bib']['end'] - new_entry['bib']['start'] < 0:
                    continue
                bib[new_entry['i']] = new_entry['bib']
            for x in set(to_del):
                del bib[x]

            # check again not overlapping dups - so if everything has worked
            ses = sorted([(i,i+x['end']-x['start']) for i,x in bib.items() if 'dup' not in x])
            starts = [x[0] for x in ses]
            ends = [x[1] for x in ses]

            for i,d in dups.items():
                end = i + d['end']-d['start']
                end_bigger_start_of_dup = bisect_left(ends,i)
                start_smaller_end_of_dup = bisect_left(starts,end)
                for idx in range(end_bigger_start_of_dup-10,min(len(ends),start_smaller_end_of_dup+10)):
                    if starts[idx] == i:
                        continue
                    if overlaps((starts[idx],ends[idx]),(i,end)):
                        print(starts[idx],ends[idx],i,end)
                        pprint(bib[starts[idx]])
                        pprint(d)
                        print('error check non overlapping dups - candidates')
                        exit(1)
        else:
            print(f'No dups file found for {org}, skipping dups processing')

        # check again general overlap...
        ses = sorted([(i,i+x['end']-x['start']) for i,x in bib.items()])
        for enu,se in enumerate(ses[:-1]):
            if se[1] >= ses[enu+1][0]:
                print('error overlapping candidates after adding dups',se,ses[enu+1])
                exit(1)

        print(len(bib))
        with open('../anchors/candidates' + f'/{org}','wb') as f:
            pickle.dump(bib,f)
